/*
  Module for the purpose of compiling a parse tree
  Copyright (C) 1999  Hardy Kahl

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

// ---------------------------------------------------------------------------- 
//  Description      : Module for the purpose of compiling a parse tree 
// ---------------------------------------------------------------------------- 
//  Remarks          : 
//    Achim     31.01.1999 
// ---------------------------------------------------------------------------- 
//  Modifications    : 
//    Achim     31.01.1999      created. 
//    Achim     01.02.1999      after deep changes in design, 
//				redifinition of interface. 
//    Philipp   06.02.1999      after some minor changes in design, 
//			        redifinition of interface (xVisitor). 
//    Achim     09.02.1999      changes were made to xIdentifier, xArray
//                              xMemoryObject was added.
// ---------------------------------------------------------------------------- 




#ifndef COMPPANO_H
#define COMPPANO_H




#include "base.h"
#include "ctlist.h"
#include "xstring.h"
#include "prgram.h"




class xParameter;
class xNode;
class xFunction;
class xCondition;
class xIdentifier;
class xVariable;
class xCommand;
class xVisitor;
class xCommandBlock;
class xExpression;




typedef xList<xVariable *>  xArgList;
typedef xList<xCommand *>   xCommandList;
  


 
struct xNode 
{ 
  TIndex ResultAddress;

  xNode();
  virtual ~xNode();

  virtual void AcceptVisitor(xVisitor &visitor) = 0; 
  virtual void print()
    { cout << "XNODE is not printable!" << endl; }
}; 
 
 
 
 
struct xFunction : public xNode 
{ 
  xVariable   *Name; 
  xParameter  *ArgList; 
  xCommand    *Command; 
 
  xFunction(xVariable *name, xParameter *arglist, xCommand *command); 
  virtual ~xFunction();

  virtual void AcceptVisitor(xVisitor &visitor); 
  virtual void print();
}; 
 
 

 
struct xParameter : public xNode 
{ 
  xArgList *ArgList; 
 
  xParameter(xArgList *arglist); 
  virtual ~xParameter();

  virtual void AcceptVisitor(xVisitor &visitor); 
  virtual void print();
}; 
 
 

 
struct xCondition : public xNode
{ 
  xExpression   *Expression; 
 
  xCondition(xExpression *expression);
  virtual ~xCondition();

  virtual void AcceptVisitor(xVisitor &visitor); 
  virtual void print();
}; 
 
 
 
 
struct xExpression : public xNode 
{
  virtual void print()
    { cout << "XEXPRESSION is not printable!" << endl; }
}; 
 
 
 
 
struct xCommand : public xNode 
{ 
  virtual void print()
    { cout << "XCOMMAND is not printable!" << endl;}
}; 
 
 
 
 
struct xCommandBlock : public xCommand 
{ 
  xCommandList  *CommandList; 
 
  xCommandBlock(xCommandList *commandlist); 
  virtual ~xCommandBlock();

  virtual void AcceptVisitor(xVisitor &visitor); 
  virtual void print();
}; 
 
 
 

struct xAssignment : public xCommand 
{ 
  xIdentifier   *Destination; 
  xExpression   *Source; 
 
  xAssignment(xIdentifier *destination, xExpression *source); 
  virtual ~xAssignment();

  virtual void AcceptVisitor(xVisitor &visitor); 
  virtual void print();
}; 
 
 
 
 
struct xWhile : public xCommand 
{ 
  xCondition            *Condition; 
  xCommand              *Command; 
 
  xWhile(xCondition *condition, xCommand *command); 
  virtual ~xWhile();

  virtual void AcceptVisitor(xVisitor &visitor);
  virtual void print(); 
}; 
 
 
 
 
struct xSum : public xExpression 
{ 
  xExpression   *Left; 
  xExpression   *Right; 
 
  xSum(xExpression *left, xExpression *right); 
  virtual ~xSum();

  virtual void AcceptVisitor(xVisitor &visitor);
  virtual void print();
}; 
 
 
 
 
struct xDiff : public xExpression 
{ 
  xExpression   *Left; 
  xExpression   *Right; 
 
  xDiff(xExpression *left, xExpression *right); 
  virtual ~xDiff();

  virtual void AcceptVisitor(xVisitor &visitor); 
  virtual void print();
}; 
 
 
 
 
struct xConstant : public xExpression 
{ 
  const xString         Value; 
 
  xConstant(xString value);
  virtual ~xConstant();

  virtual void AcceptVisitor(xVisitor &visitor); 
  virtual void print();
}; 




struct xMemoryObject
{
 public:
  TIndex* Address;
  xMemoryObject();
  ~xMemoryObject();
};
 
 
 
 
enum EVarType 
{
  Unknown,
  Function,
  Simple,  
  Array
};




struct xIdentifier : public xExpression 
{ 
  xString        Name;
  xMemoryObject* Address;
  EVarType       Type;
  
  xIdentifier(xString name);
  virtual ~xIdentifier();
  virtual void print()
    { cout << "IDENTIFIER is not printable!" << endl; }
}; 
  



struct xVariable : public xIdentifier
{
  xVariable(xString name);
  virtual ~xVariable();

  virtual void AcceptVisitor(xVisitor &visitor);
  virtual void print();
  TBool operator==(xVariable variable);
};



 
struct xArray : public xIdentifier 
{
  xExpression   *Index; 
 
  xArray(xString name, xExpression *index);
  virtual ~xArray();

  virtual void AcceptVisitor(xVisitor &visitor); 
  virtual void print();
  TBool operator==(xArray array); 
}; 




typedef xList<TIndex> xRegList;




class xMemoryManager
{
 public:
  void Free(TIndex RegisterNo);       
  TIndex GetAddress(xIdentifier& Node);
  TIndex GetTmpReg();
  xMemoryManager(TIndex noofvar, TIndex noofarray, TIndex baseaddress);

 private:
  xRegList    FreedTmpReg;
  TIndex      LastUsedTmpRegister;
  TIndex      LastUsedArrayRegister;
  TIndex      LastUsedVarRegister;
  TIndex      PageSize;
};




class xVisitor 
{ 
 public:
  void visit_xFunction(xFunction *Node);
  void visit_xParameter(xParameter *Node);
  void visit_xCommandBlock(xCommandBlock *Node);
  void visit_xWhile(xWhile *Node);
  void visit_xCondition(xCondition *Node);
  void visit_xAssignment(xAssignment *Node);
  void visit_xExpression(xExpression *Node);
  void visit_xVariable(xVariable *Node);
  void visit_xArray(xArray *Node);
  void visit_xConstant(xConstant *Node);
  void visit_xSum(xSum *Node);
  void visit_xDiff(xDiff *Node);
  xVisitor(TIndex no_of_args, TIndex no_of_var, TIndex no_of_array, xFunction *main_function, xString filename);

 protected:
 
 private:
  xRAMProgram     DestProg;              // this is the destination programm
  TIndex          CCL;                   // this variable is to count the current line in the destination program DestProg,
                                         // the line where to insert new code
  xString         Zero;                  // this variable hold the name of the register loaded with 0. This register is used if
                                         // the compiler neads to calculate something. e.g. array-indexes
  xString         One;                   // look at Zero. In contrast to Zero this Register is loaded with 1
  xString         PageSize;               
  xString         Counter;               // this Register is used when the compiler nead a loopcounter. e.g. array-indexes
  xMemoryManager  MemManager;            // an object which manages memoryallocation
  xString         FileName;
			  
  xString Index2Register(TIndex regno); // this methode converts a TIndex an an xString of the form "R" + TIndex
  xString InsertCCL();                  // this mehtode returns CCL 
}; 




#endif



