/*
  List container
  Copyright (C) 1999  Hardy Kahl

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

// ----------------------------------------------------------------------------
//  Description      : List container
// ----------------------------------------------------------------------------
//  Remarks          :
//    The iterator functions return the current element and reposition
//    afterwards. (except pre-increments)
//    The operator~() function returns the number of elements still available.
//
// ----------------------------------------------------------------------------
//  (c) Copyright 1996 by iXiONmedia, all rights reserved.
// ----------------------------------------------------------------------------




#ifndef CTLIST_H
#define CTLIST_H




#include <assert.h>
#include "base.h"
#include "numeric.h"
#include "ctbase.h"
#include "auto.h"




// xListBase ------------------------------------------------------------------
template <class T>
class xListBase : public xContainer<T> {
  protected:
  xArrayAllocator<T> List;
  TSize Items;
  TSize DeltaSize;

  public:
  xListBase(TSize size = 50,TSize deltasize = 10)
    : List(size), Items(0), DeltaSize(deltasize)
    { }
  virtual ~xListBase()
    { Empty(); }
  void Optimize()
    { Resize(Items); }
  inline void Empty();

  TSize operator~() const
    { return Items; }
  inline void operator-=(TIndex index);
  T &operator[](TIndex index) const
    {
      if (index >= Items) EX_THROW(Generic,EC_INDEX);
      return List[index];
    }

  protected:
  inline TIndex InternalInsert(TIndex index,T object);
  private:
  inline void Resize(TSize newsize);
};




// xList ----------------------------------------------------------------------
template <class T>
struct xList : public xListBase<T> {
  xList(TSize size = 50,TSize deltasize = 10)
    : xListBase<T>(size,deltasize)
    { }

  TIndex InsertFirst(T object)
    { return InternalInsert(0,object); }
  TIndex InsertLast(T object)
    { return InternalInsert(Items,object); }
  TIndex operator+=(T object)
    { return InternalInsert(Items,object); }
  TIndex InsertPos(TIndex index,T object)
    { return InternalInsert(index,object); }
};




// xSearchableList ------------------------------------------------------------
template <class T>
struct xSearchableList : public xList<T> {
  xSearchableList(TSize size = 50,TSize deltasize = 10)
    : xList<T>(size,deltasize)
    { }
  inline TIndex operator()(T &object,TIndex start = 0);
};




// xSortedList ----------------------------------------------------------------
template <class T>
class xSortedList : public xListBase<T>,protected xComparator<T> {
  public:
  xSortedList(TSize size = 50,TSize deltasize = 10)
    : xListBase<T>(size,deltasize)
    { }

  inline TIndex operator+=(T object);
  inline TIndex operator()(T &object,TIndex start = 0);
};




// xListIterator --------------------------------------------------------------
template <class T>
class xListIterator {
  TIndex BaseIndex;
  TIndex Index;
  xListBase<T> &List;

  public:
  xListIterator(xListBase<T> &list,TIndex baseindex = 0)
    : BaseIndex(baseindex),Index(baseindex),List(list)
    { };
  void Reset()
    { Index = BaseIndex; }
  void Reset(TIndex baseindex)
    {
      BaseIndex = baseindex;
      Index = BaseIndex;
    }
  TIndex operator!()
    { return Index; }
  TSize operator~()
    { return Index >= ~List ? 0 : ~List-Index; }
  T &operator++(int)
    { return List[Index++]; }
  T &operator--(int)
    { return List[Index--]; }
  T &operator++()
    { return List[++Index]; }
  T &operator--()
    { return List[--Index]; }
  T &operator+=(TDeltaD delta)
    { return List[(Index += delta) - delta]; }
  T &operator-=(TDeltaD negdelta)
    { return List[(Index -= negdelta) + negdelta]; }
  T &operator+(TDeltaD delta)
    { return List[Index + delta]; }
  T &operator-(TDeltaD negdelta)
    { return List[Index - negdelta]; }
  T &operator()()
    { return List[Index]; }
};




// xListBase ------------------------------------------------------------------
template <class T>
void xListBase<T>::Empty()
{
  for (TIndex itemcount = 0;itemcount<Items;itemcount++)
    Kill(List[itemcount]);
  Items = 0;
}




template <class T>
void xListBase<T>::operator-=(TIndex index)
{
  if (index>=Items)
    throw xContainerException(ECCT_INDEX,"delete");
  Kill(List[index]);
  for (TIndex item = index+1;item<Items;item++)
    List[item-1] = List[item];
  Items--;
}




template <class T>
TIndex xListBase<T>::InternalInsert(TIndex index,T object)
{
  if (index>Items) throw xContainerException(ECCT_INDEX,"insert");
  if (Items==~List) Resize(~List+DeltaSize);
  for (TIndex item = Items;item>index;item--)
    List[item] = List[item-1];
  List[index] = object;
  Create(List[index]);
  Items++;

  return index;
}




template <class T>
void xListBase<T>::Resize(TSize newsize)
{
  if (newsize<Items) throw xContainerException(ECCT_RESIZE);
  List(newsize);
}




// xSearchableList ------------------------------------------------------------
template <class T>
TIndex xSearchableList<T>::operator()(T &object,TIndex start)
{
  for (TIndex itemcount = start;itemcount<Items;itemcount++)
    if (object == List[itemcount]) return itemcount;
  throw xContainerException(ECCT_NOTFOUND);
}




// xSortedList ----------------------------------------------------------------
template <class T>
TIndex xSortedList<T>::operator+=(T object)
{
  TIndex previndex = 0;
  TIndex index = 0;
  TDelta jumpsize = Items / 2;

  forever {
    if (index == 0 || Compare(object,List[index-1]) >= REqual) {
      if (index >= Items || Compare(object,List[index]) <= REqual) {
        InternalInsert(index,object);
        return index;
      }
      else previndex = index;
    }
    else index = previndex;

    index += jumpsize;
    jumpsize /= 2;
    if (!jumpsize) jumpsize = 1;
  }
}




template <class T>
TIndex xSortedList<T>::operator()(T &object,TIndex start)
{
  for (TIndex itemcount = start;itemcount<Items;itemcount++)
    if (Compare(object,List[itemcount])==REqual) return itemcount;
  throw xContainerException(ECCT_NOTFOUND);
}




#endif
