/*
  Random Access Machine Simulator
  Copyright (C) 1999  Hardy Kahl

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

// ----------------------------------------------------------------------------
//  Description      : Random Access Machine Simulator.
// ----------------------------------------------------------------------------




#include "ram.h"
#include "bcode.h"
#include "compram.h"
#include <string.h>




// Plain text rendering table -------------------------------------------------
#define T_INVALID_BCODE         "Invalid byte code"
#define T_INVALID_GOTO          "Invalid jump address"




static char *(PlainText[]) =
  { T_INVALID_BCODE, T_INVALID_GOTO };




// xRAMException --------------------------------------------------------------
xRAMException::xRAMException(TErrorCode error, TIndex line, xString &info)
: xBaseException(error, NULL, NULL, 0, "RAM")
{
  HasInfo = TRUE;
  try {
    xString temp = "In line ";
    temp += xString::Unsigned2Dec(line + 1);
    temp += " : ";
    temp += info;
    strcpy(Info, temp);
  }
  catch (...) { }
}




char *xRAMException::GetText() const
{
  return PlainText[Error];
}




// xRAM : ---------------------------------------------------------------------
xRAM::xRAM(void)
: CommandRegister(0), Terminated(FALSE)
{
}




void xRAM::Load(xRAMProgram &program)
{
  BCProgram.Empty();
  CompileRAM2BC(program, BCProgram);
  Reset();
}




void xRAM::Reset(void)
{
  CommandRegister = 0;
  Terminated = FALSE;
  RegisterManager.RegisterList.Empty();
}




void xRAM::Execute(void)
{
  while (!Terminated) Execute(1);
}




void xRAM::Execute(TCount numcmd)
{
  if (CommandRegister < ~BCProgram) {
    xByteCode bcode = BCProgram[CommandRegister];
    switch (bcode.Code) {
      case BC_STOP:
        Terminated = TRUE;
        break;
      case BC_MOV_R_R:
        RegisterManager[bcode.Param1] = RegisterManager[bcode.Param2];
        CommandRegister ++;
        break;
      case BC_MOV_R_RR:
        RegisterManager[bcode.Param1] = RegisterManager[RegisterManager[bcode.Param2]];
        CommandRegister ++;
        break;
      case BC_MOV_RR_R:
        RegisterManager[RegisterManager[bcode.Param1]] = RegisterManager[bcode.Param2];
        CommandRegister ++;
        break;
      case BC_MOV_R_K:
        RegisterManager[bcode.Param1] = bcode.Param2;
        CommandRegister ++;
        break;
      case BC_ADD:
        RegisterManager[bcode.Param1] = RegisterManager[bcode.Param2] + RegisterManager[bcode.Param3];
        CommandRegister ++;
        break;
      case BC_SUB:
        RegisterManager[bcode.Param1] = RegisterManager[bcode.Param2] - RegisterManager[bcode.Param3];
        CommandRegister ++;
        break;
      case BC_GOTO:
        try {
          CommandRegister = bcode.Param1.EvalUnsigned() - 1;
        }
        catch (...) {
          throw xRAMException(ECRAM_INVALID_GOTO, CommandRegister, bcode.Param1);
        }
        break;
      case BC_IF_NULL:
        if (RegisterManager[bcode.Param1].IsZero()) {
          try {
            CommandRegister = bcode.Param2.EvalUnsigned() - 1;
          }
          catch (...) {
            throw xRAMException(ECRAM_INVALID_GOTO, CommandRegister, bcode.Param2);
          }
        }
        else CommandRegister ++;
        break;
      case BC_IF_NOT_NULL:
        if (!(RegisterManager[bcode.Param1].IsZero())) {
          try {
            CommandRegister = bcode.Param2.EvalUnsigned() - 1;
          }
          catch (...) {
            throw xRAMException(ECRAM_INVALID_GOTO, CommandRegister, bcode.Param2);
          }
        }
        else CommandRegister ++;
        break;
      default:
        xString codestr;
        codestr.Unsigned2DecThis(bcode.Code);
        throw xRAMException(ECRAM_INVALID_BCODE, CommandRegister, codestr);
        break;
    }
  }
  if (CommandRegister >= ~BCProgram) {
    xString empty = "";
    throw xRAMException(ECRAM_INVALID_BCODE, CommandRegister, empty);
  }
}




void xRAM::ExecuteTo(TIndex line)
{
  while ((CommandRegister != line) && !Terminated) Execute(1);
}
