/*
  Regular expressions string object.
  Copyright (C) 1999  Hardy Kahl

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

// ----------------------------------------------------------------------------
//  Description      : Regular expressions string object.
// ----------------------------------------------------------------------------
//  (c) Copyright 1998 by iXiONmedia, all rights reserved.
// ----------------------------------------------------------------------------




#include <ctype.h>
#include "exgen.h"
#include "xstrre.h"




// Error texts ----------------------------------------------------------------
#define TRE_INVQUANTIFIER             "Invalid quantifier"
#define TRE_UNBALBACKREF              "Unbalanced backreference"
#define TRE_INVESCAPE                 "Invalid escape sequence"
#define TRE_INVBACKREF                "Invalid backreference"
#define TRE_UNTERMCLASS               "Unterminated character class"




static char *RegexPlainText[] = {
  TRE_INVQUANTIFIER,TRE_UNBALBACKREF,TRE_INVESCAPE
};




// xRegexException ------------------------------------------------------------
xRegexException::xRegexException(TErrorCode error,
  char const *info,char *module,TIndexD line)
: xBaseException(error,info,module,line,"RE")
{
}




char *xRegexException::GetText() const
{
  return RegexPlainText[Error];
}




// xRegexMatcher --------------------------------------------------------------
TSize xRegexMatcher::CollectMatchLength() const
{
  TSize totalml = 0;
  xRegexMatcher const *object = this;
  while (object) {
    totalml += object->MatchLength;
    object = object->Next;
  }
  return totalml;
}




// xRegexQuantifier -----------------------------------------------------------
TBool xRegexQuantifier::Match(xRegexBackrefHeap &brheap,
  xString const &candidate,TIndex at)
{
  TSize matchcount = 0;
  if (Greedy) {
    matchcount = ~candidate - at;
    TSize prevmatchcount;
    while ((prevmatchcount = matchcount--))
      if (MatchCount(brheap,candidate,at,prevmatchcount)) return YES;
  }
  else {
    TSize maxmatchcount = ~candidate - at;
    while (matchcount <= maxmatchcount) {
      if (MatchCount(brheap,candidate,at,matchcount)) return YES;
      matchcount++;
    }
  }
  return NO;
}




TBool xRegexQuantifier::MatchCount(xRegexBackrefHeap &brheap,
  xString const &candidate,TIndex at,TSize matchcount)
{
  if (!InRange(matchcount)) return NO;

  MatchLength = 0;
  while (matchcount--) {
    if (!MatchQuantified(brheap,candidate,at+MatchLength)) return NO;
    MatchLength += GetQuantifiedMatchLength();
  }

  if (!MatchNext(brheap,candidate,at+MatchLength)) return NO;
  return YES;
}




// xRegexClassMatcher ---------------------------------------------------------
void xRegexClassMatcher::ExpandClass()
{
  if (!~CharClass) return;
  xString finalclass = CharClass[0];
  char lastchar = CharClass[0];

  for (TIndex index = 1;index < ~CharClass;index++) {
    if ((CharClass[index] == XSTRRE_CLASSRANGE) && (index < ~CharClass-1)) {
      for (char ch = lastchar;ch < CharClass[index+1];ch++)
        CharClass += ch;
    }
    else finalclass += CharClass[index];
  }
  CharClass = finalclass;
}




// xRegexBackrefOpenMatcher ---------------------------------------------------
TBool xRegexBackrefOpenMatcher::Match(xRegexBackrefHeap &brheap,
  xString const &candidate,TIndex at)
{
  xRegexBackref backref;
  backref.Checkpoint(at);

  TSize brindex = brheap += backref;

  TBool result = MatchNext(brheap,candidate,at);

  if (!result) { brheap -= brindex; }
  return result;
}




// xRegexBackrefCloseMatcher --------------------------------------------------
TBool xRegexBackrefCloseMatcher::Match(xRegexBackrefHeap &brheap,
  xString const &candidate,TIndex at)
{
  if (!~brheap) EX_THROW(Regex,ECRE_UNBALBACKREF)
  TIndex last = ~brheap-1;
  brheap[last].MakeReady(at);
  return MatchNext(brheap,candidate,at);
}




// xRegexAlternativeMatcher ---------------------------------------------------
xRegexAlternativeMatcher::~xRegexAlternativeMatcher()
{
  while (~AltList) {
    xRegexMatcher *matcher = AltList[0];
    AltList -= 0;
    delete matcher;
  }
}




void xRegexAlternativeMatcher::operator+=(xRegexMatcher *matcher)
{
  AltList += matcher;
  xRegexMatcher *searchlast = matcher,*last = NULL;
  while (searchlast) {
    last = searchlast;
    searchlast = searchlast->GetNext();
  }
  last->SetNext(&Connector,NO);
}




TBool xRegexAlternativeMatcher::Match(xRegexBackrefHeap &brheap,
  xString const &candidate,TIndex at)
{
  xListIterator<xRegexMatcher *> it(AltList);
  while (~it) {
    if (it()->Match(brheap,candidate,at)) {
      MatchLength = 0;
      xRegexMatcher const *object = it();
      while (object != &Connector) {
        MatchLength += object->GetMatchLength();
        object = object->GetNext();
      }
      return YES;
    }
    it++;
  }
  return NO;
}




// xRegexBackrefMatcher -------------------------------------------------------
TBool xRegexBackrefMatcher::Match(xRegexBackrefHeap &brheap,
  xString const &candidate,TIndex at)
{
  xString matchstr = brheap[Backref].Get(candidate);
  MatchLength = ~matchstr;

  if (at+~matchstr > ~candidate) return NO;
  return (candidate(at,~matchstr) == matchstr) &&
    MatchNext(brheap,candidate,at+~matchstr);
}




// xRegexString ---------------------------------------------------------------
TBool xRegexString::Match(xString const &candidate,TIndex from)
{
  LastCandidate = candidate;
  BackrefHeap.Empty();
  if (!ParsedRegex) return YES;
  for (TIndex index = from;index < ~candidate;index++)
    if (ParsedRegex->Match(BackrefHeap,candidate,index)) {
      MatchIndex = index;
      MatchLength = ParsedRegex->CollectMatchLength();
      return YES;
    }
  return NO;
}




TBool xRegexString::MatchAt(xString const &candidate,TIndex at)
{
  LastCandidate = candidate;
  BackrefHeap.Empty();
  if (!ParsedRegex) return YES;
  if (ParsedRegex->Match(BackrefHeap,candidate,at)) {
    MatchIndex = at;
    MatchLength = ParsedRegex->CollectMatchLength();
    return YES;
  }
  return NO;
}




xString xRegexString::ReplaceAll(xString const &candidate,
  xString const &replacement,TIndex from)
{
  xString result;
  xString tempreplacement;

  LastCandidate = candidate;
  BackrefHeap.Empty();

  if (!ParsedRegex) return candidate;

  for (TIndex index = from;index < ~candidate;) {
    if (ParsedRegex->Match(BackrefHeap,candidate,index)) {
      TIndex matchlength = ParsedRegex->CollectMatchLength();
      tempreplacement = replacement;

      xRegexBackrefHeapIterator it(BackrefHeap);
      while (~it) {
        tempreplacement.ReplaceThis(XSTRRE_BACKREF+xString::Unsigned2Dec(!it),
          BackrefHeap[!it].Get(LastCandidate));
        it++;
      }

      result += tempreplacement;
      index += matchlength;
    }
    else result += candidate[index++];
  }
  return result;
}




xRegexMatcher *xRegexString::ParseRegex(xString const &expr)
{
  if (!~expr) return NULL;
  TIndex index = 0;
  xRegexMatcher *firstobject = NULL;
  xRegexMatcher *lastobject  = NULL;
  xRegexAlternativeMatcher *alternative = NULL;

  while (index < ~expr) {
    xRegexMatcher *object = NULL;
    xRegexQuantifier *quantifier = NULL;
    TBool quantified = YES;

    // several objects may be inserted in one loop run

    switch (expr[index++]) {
      // case XSTRRE_SUBPATTERN:
      // case XSTRRE_ESCAPESEQ:
      case XSTRRE_LITERAL: {
          if (index >= ~expr) EX_THROW(Regex,ECRE_INVESCAPE)

          char ch = expr[index++];
          if (isdigit(ch))
            object = new xRegexBackrefMatcher(ch-'0');
          else
            object = new xRegexStringMatcher(ch);
          EX_MEMCHECK(object)
          break;
        }
      case XSTRRE_ANYCHAR:
        object = new xRegexAnyCharMatcher;
        EX_MEMCHECK(object)
        break;
      case XSTRRE_START:
        quantified = NO;
        object = new xRegexStartMatcher;
        EX_MEMCHECK(object)
        break;
      case XSTRRE_END:
        quantified = NO;
        object = new xRegexEndMatcher;
        EX_MEMCHECK(object)
        break;
      case XSTRRE_ALTERNATIVE: {
          if (!alternative) {
            alternative = new xRegexAlternativeMatcher;
            EX_MEMCHECK(alternative)
          }
          (*alternative) += firstobject;
          firstobject = NULL;
          lastobject = NULL;
          break;
        }
      case XSTRRE_CLASSSTART:
        try {
          TIndex classend = expr.Locate(XSTRRE_CLASSSTOP,index);
          object = new xRegexClassMatcher(expr(index,classend-index));
          EX_MEMCHECK(object)

          index = classend+1;
        }
        EX_CONVERT(Generic,EC_ITEMNOTFOUND,Regex,ECRE_UNTERMCLASS)
        break;
      case XSTRRE_BACKREFSTART: {
          quantified = NO;
          xRegexMatcher *parsed = NULL;

          TSize brlevel = 1;
          for (TIndex searchstop = index;searchstop < ~expr;searchstop++) {
            if ((expr[searchstop] == XSTRRE_BACKREFSTART) &&
            (expr[searchstop-1] != XSTRRE_LITERAL))
              brlevel++;
            if ((expr[searchstop] == XSTRRE_BACKREFSTOP) &&
            (expr[searchstop-1] != XSTRRE_LITERAL)) {
              brlevel--;
              if (brlevel == 0) {
                parsed = ParseRegex(expr(index,searchstop-index));
                if (!parsed) EX_THROW(Regex,ECRE_INVBACKREF)

                index = searchstop+1;
                break;
              }
            }
          }

          if (!parsed) EX_THROW(Regex,ECRE_UNBALBACKREF)

          object = new xRegexBackrefOpenMatcher;
          EX_MEMCHECK(object)
          object->SetNext(parsed);

          xRegexMatcher *closer = new xRegexBackrefCloseMatcher;
          EX_MEMCHECK(closer);

          xRegexMatcher *searchlast = parsed;
          xRegexMatcher *foundlast  = NULL;
          while (searchlast) {
            foundlast = searchlast;
            searchlast = searchlast->GetNext();
          }
          foundlast->SetNext(closer);

          break;
        }
      case XSTRRE_BACKREFSTOP:
        EX_THROW(Regex,ECRE_UNBALBACKREF)
      default:
        object = new xRegexStringMatcher(expr[index-1]);
        EX_MEMCHECK(object)
        break;
    }

    if (object) {
      if (quantified) quantifier = ParseQuantifier(expr,index);
      if (quantifier) {
        quantifier->SetQuantified(object);
        if (lastobject) lastobject->SetNext(quantifier);
        else firstobject = quantifier;
      }
      else {
        if (lastobject) lastobject->SetNext(object);
        else firstobject = object;
      }
    }

    xRegexMatcher *searchlast = quantifier ? quantifier : object;
    while (searchlast) {
      lastobject = searchlast;
      searchlast = searchlast->GetNext();
    }
  }
  if (alternative) {
    (*alternative) += firstobject;
    return alternative;
  }
  else return firstobject;
}




xRegexQuantifier *xRegexString::ParseQuantifier(xString const &expr,TIndex &at)
{
  xRegexQuantifier *quant = NULL;
  if (at == ~expr) return NULL;
  if (expr[at] == XSTRREQ_0PLUS) {
    quant = new xRegexQuantifier(IsGreedy(expr,++at),0);
    EX_MEMCHECK(quant)
    return quant;
  }
  if (expr[at] == XSTRREQ_1PLUS) {
    quant = new xRegexQuantifier(IsGreedy(expr,++at),1);
    EX_MEMCHECK(quant)
    return quant;
  }
  if (expr[at] == XSTRREQ_01) {
    quant = new xRegexQuantifier(IsGreedy(expr,++at),0,1);
    EX_MEMCHECK(quant)
    return quant;
  }
  if (expr[at] == XSTRREQ_START) {
    TSize min,max;

    at++;
    TIndex endindex;
    xString quantspec;
    try {
      endindex = expr.Locate(XSTRREQ_STOP,at);
      quantspec = expr(at,endindex-at);
    }
    EX_CONVERT(Generic,EC_ITEMNOTFOUND,Regex,ECRE_INVQUANTIFIER)

    at = endindex+1;

    try {
      try {
        TIndex rangeindex = quantspec.Locate(XSTRREQ_RANGE);
        if (rangeindex == ~quantspec-1) {
          min = quantspec.Delete(rangeindex,1).EvalUnsigned();
          quant = new xRegexQuantifier(IsGreedy(expr,at),min);
        }
        else {
          min = (quantspec << rangeindex).EvalUnsigned();
          max = quantspec(rangeindex+1).EvalUnsigned();
          quant = new xRegexQuantifier(IsGreedy(expr,at),min,max);
        }
      }
      EX_CATCHCODE(Generic,EC_ITEMNOTFOUND,
        min = quantspec.EvalUnsigned();
        quant = new xRegexQuantifier(IsGreedy(expr,at),min,min);
      )
      EX_MEMCHECK(quant);
      return quant;
    }
    EX_CONVERT(Generic,EC_CANNOTEVALUATE,Regex,ECRE_INVQUANTIFIER)
  }
  return NULL;
}




TBool xRegexString::IsGreedy(xString const &expr,TIndex &at)
{
  if (at == ~expr) return YES;
  if (expr[at] == XSTRREQ_NONGREEDY) {
    at++;
    return NO;
  }
  else return YES;
}

