/*
  Module for the purpose of compiling a parse tree
  Copyright (C) 1999  Philipp Cimiano Lavin

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

// ---------------------------------------------------------------------------- 
//  Description      : Module for the purpose of compiling a parse tree 
// ---------------------------------------------------------------------------- 
//  Remarks          : 
// ---------------------------------------------------------------------------- 
//  Modifications    : 
//      Phil         02.07.1999  created    
//
// --------------------------------------------------------------------------


#include "frontend.h"
#include "backend.h"
#include <iostream.h>

//xNode:
//no overloading of AcceptVisitor!!!

xNode::xNode()
{
}

xNode::~xNode()
{
}



//xConstant

xConstant::xConstant(xString value)
: Value(value) 
{
}

xConstant::~xConstant()
{
}

void xConstant::AcceptVisitor(xPascalliVisitor & visitor)
{
  visitor.visit_xConstant((*this));  
}

void xConstant::print()
{
  cout << Value <<" (" << Alpha << "," << Beta <<")";
}



//xIdentifier:

xIdentifier::xIdentifier(xString name)
: Name(name), Type(Unknown)
{
}

xIdentifier:: ~xIdentifier()
{
}


void xIdentifier::AcceptVisitor(xPascalliVisitor &visitor)
{
  visitor.visit_xIdentifier((*this));
}

void xIdentifier::print()
{
  cout << Name  <<" (" << Alpha << "," << Beta <<")";
}

TBool xIdentifier::operator==(xIdentifier identifier)
{
  return (Name==(identifier.Name));
}


//xVariable:

xVariable::xVariable(xString name)
:xIdentifier(name)
{
  Type = Simple;
}

xVariable::~xVariable()
{
}

void xVariable::AcceptVisitor(xPascalliVisitor &visitor)
{
  visitor.visit_xIdentifier((*this));
}


void xVariable::print()
{
  cout << Name  <<" (" << Alpha << "," << Beta <<")";
}


//xArray:

xArray::xArray(xString name, xExpression *index)
:xIdentifier(name), Index(index)
{
  Type = Array;
}

xArray::~xArray()
{
  delete(Index);
  Index=0;
}

void xArray::AcceptVisitor(xPascalliVisitor &visitor)
{
  visitor.visit_xArray((*this));
}


void xArray::print()
{
  cout << Name <<"(";
  if (Index)
  Index->print();
  cout <<")"  <<" (" << Alpha << "," << Beta <<")";
}


//xFunctionCall:

xFunctionCall::xFunctionCall(xString name, xExpressionList &arguments)

: xIdentifier(name), Arguments(arguments)
{
  Type = Function;
}

xFunctionCall::~xFunctionCall() 
{
  for (TIndex i=0; i<~Arguments; i++)
    {
      delete(Arguments[i]);
    }
}

void xFunctionCall::AcceptVisitor(xPascalliVisitor &visitor)
{
  visitor.visit_xFunctionCall((*this));
}


void xFunctionCall::print()
{
  cout <<Name <<"(";
  for (TIndex i=0; i<~Arguments; i++)
  {
    Arguments[i]->print();
    cout << ",";
  }
  cout <<")"  <<" (" << Alpha << "," << Beta <<")";
} 

//xProgramNode:

xProgramNode::xProgramNode(xFunctionList &functions)
: Functions(functions)
{
}

xProgramNode::~xProgramNode()
{
  for (TIndex i=0; i<~Functions; i++)
    {
      delete(Functions[i]);
    }
}


void xProgramNode::AcceptVisitor(xPascalliVisitor &visitor)
{
  visitor.visit_xProgramNode((*this));
}


void xProgramNode::print()
{
  cout << "printing program" <<endl;
 for (TIndex i=0; i<~Functions; i++)
   {
     Functions[i]->print();
   } 
}


//xFunction:

xFunction::xFunction(xIdentifier *name, xIdentifierList &arguments, xCommand *command)
: Name((*name)), Arguments(arguments), Command(command)
{
}

xFunction:: ~xFunction()
{
  for (TIndex i=0; i<~Arguments; i++)
    {
      delete(Arguments[i]);
    }
  delete(Command);
}

void xFunction::AcceptVisitor(xPascalliVisitor &visitor)
{
  visitor.visit_xFunction((*this));
}

void xFunction::print()
{
  cout <<"FUNCTION" << " " <<Name.Name <<"(";
  for (TIndex i=0; i<~Arguments ;i++)
    {
      Arguments[i]->print();
      cout << ",";
    }
  cout <<");\n";
  if (Command)
    {
      Command->print();
    }
}


//xCommandBlock:

xCommandBlock::xCommandBlock(xCommandList &commandlist)
  : CommandList(commandlist)
{
};




xCommandBlock::xCommandBlock()
  : CommandList(*(new xCommandList))
{
};




xCommandBlock::~xCommandBlock()
{
  for (TIndex i=0; i<~CommandList; i++)
  {
    delete(CommandList[i]);
  };
};




void xCommandBlock::
addCommand(xCommand* newcommand)
{
  CommandList += newcommand;
};




xCommand* xCommandBlock::
getLastCommand()
{
  return CommandList[~CommandList -1];
};




void xCommandBlock::AcceptVisitor(xPascalliVisitor &visitor)
{
  visitor.visit_xCommandBlock((*this));
}


void xCommandBlock::print()
{
  cout <<"BEGIN" <<endl;
  for (TIndex i=0; i<~CommandList ;i++)
    {
      CommandList[i]->print();
    }
  cout <<"END"  <<" (" << Alpha << "," << Beta <<")";
  cout <<endl;
}


TBool xCommandBlock::variable_used(xIdentifier *identifier)
{
  for (TIndex i=0; i<~CommandList; i++)
    {
      if (CommandList[i]->variable_used(identifier))
	{
	  return TRUE;
	}
    }
  return FALSE;
}
 
//xAssignment:

xAssignment::xAssignment(xIdentifier *destination, xExpression *source)
  : Destination(destination), Source(source)
{
}

xAssignment::~xAssignment()
{
  delete(Destination);
  Destination=0;
  delete(Source);
  Source=0;
}

void xAssignment::AcceptVisitor(xPascalliVisitor &visitor)
{
  visitor.visit_xAssignment((*this));
}


void xAssignment::print()
{
  if (Destination)
    {
      Destination->print();
    }
  cout <<" : = ";
  if (Source)
    {
      Source->print();
    }
  cout  <<" (" << Alpha << "," << Beta <<")" <<endl;
}


TBool xAssignment::variable_used(xIdentifier *identifier)
{
  if ((*Destination)==(*identifier))
    {
      return TRUE;
    }
  return FALSE;
}



//xWhile:

xWhile::xWhile(xCondition *condition, xCommand *command)
:Condition(condition), Command(command) 
{
}

xWhile::~xWhile()
{
  delete(Condition);
  Condition=0;
  delete(Command);
  Command=0;
}

void xWhile::AcceptVisitor(xPascalliVisitor &visitor)
{
  visitor.visit_xWhile((*this));
}


void xWhile::print()
{
  cout <<"WHILE ";
  if (Condition)
    {
  Condition->print();
    }
  cout <<" DO ";
  cout <<endl;
  if (Command)
    {
      Command->print();
    }
  cout <<" (" << Alpha << "," << Beta <<")";
}


TBool xWhile::variable_used(xIdentifier *identifier)
{
  return Command->variable_used(identifier);
}


//xIfStatement:

xIfStatement::xIfStatement(xCondition *condition, xCommand *ifTwig, xCommand *elseTwig)
  : Condition(condition), IfTwig(ifTwig), ElseTwig(elseTwig)
{
}


xIfStatement:: ~xIfStatement()
{
  delete(Condition);
  Condition=0;
  delete(IfTwig);
  IfTwig=0;
  delete(ElseTwig);
  ElseTwig=0;
}


void xIfStatement::AcceptVisitor(xPascalliVisitor &visitor)
{
  visitor.visit_xIfStatement((*this));
}


void xIfStatement::print()
{
  cout <<"IF ";
  if (Condition)
    {
      Condition->print();
    }
  cout <<" THEN";
  cout <<endl;
  if (IfTwig)
    {
      IfTwig->print();
    }
  if (ElseTwig)
    {
      cout <<" ELSE";
      cout <<"\n";
      ElseTwig->print();
    }
   cout <<" (" << Alpha << "," << Beta <<")";
}


TBool xIfStatement::variable_used(xIdentifier *identifier)
{
  if (IfTwig->variable_used(identifier))
    {
      return TRUE;
    }
  if ((ElseTwig)&&(ElseTwig->variable_used(identifier)))
    {
      return TRUE;
    }
  return FALSE;
}




//xForLoop:


xForLoop::xForLoop(xIdentifier *counter, xExpression *start, xExpression *end,
                   xCommand *command)
  : Counter(counter), StartValue(start), EndValue(end), Command(command)
{
}


xForLoop::~xForLoop()
{
  delete(Counter);
  Counter=0;
  delete(StartValue);
  StartValue=0;
  delete(EndValue);
  EndValue=0;
  delete(Command);
  Command=0;
}



void xForLoop::AcceptVisitor(xPascalliVisitor &visitor)
{
  visitor.visit_xForLoop((*this));
}


void xForLoop::print()
{
  cout <<"FOR ";
  if (Counter)
    {
      Counter->print();
    }
  cout <<":=";
  if (StartValue)
    {
      StartValue->print();
    }
  cout <<" TO ";
  if (EndValue)
    {
  EndValue->print();
    }
  cout <<endl;
  if (Command)
    {
      Command->print();
    }
   cout <<" (" << Alpha << "," << Beta <<")";
}



TBool xForLoop::variable_used(xIdentifier *identifier)
{
  if ((*Counter)==(*identifier))
    {
      return TRUE;
    }
  if (Command->variable_used(identifier))
    {
      return TRUE;
    }
  return FALSE;
}



//xAddition:

xAddition::xAddition(xExpression *left, xExpression *right)
  : Left(left), Right(right)
{
}


xAddition::~xAddition()
{
  delete(Left);
  Left=0;
  delete(Right);
  Right=0;
}


void xAddition::AcceptVisitor(xPascalliVisitor &visitor)
{
  visitor.visit_xAddition((*this));
}


void xAddition::print()
{
  cout <<"(";
  if (Left)
    {
  Left->print();
    }
  cout <<"+";
  if (Right)
    {
  Right->print();
    }
  cout <<")"  <<" (" << Alpha << "," << Beta <<")";
}


//xSubstraction:

xSubtraction::xSubtraction(xExpression *left, xExpression *right)
: Left(left), Right(right)
{
}


xSubtraction::~xSubtraction()
{
  delete(Left);
  Left=0;
  delete(Right);
  Right=0;
}


void xSubtraction::AcceptVisitor(xPascalliVisitor &visitor)
{
  visitor.visit_xSubtraction((*this));
}

void xSubtraction::print()
{
  cout <<"(";
  if (Left)
    {
  Left->print();
    }
  cout <<"-";
  if (Right)
    {
      Right->print();
    }
  cout <<")"  <<" (" << Alpha << "," << Beta <<")";
}


//xMultiplication:

xMultiplication::xMultiplication(xExpression *left, xExpression *right)
  : Left(left), Right(right)
{
}


xMultiplication::~xMultiplication()
{
  delete(Left);
  Left=0;
  delete(Right);
  Right=0;
}


void xMultiplication::AcceptVisitor(xPascalliVisitor &visitor)
{
  visitor.visit_xMultiplication((*this));
}

void xMultiplication::print()
{
  cout <<"(";
  if (Left)
    {
  Left->print();
    }
  cout <<"*";
  if (Right)
    {
  Right->print();
    }
  cout <<")"  <<" (" << Alpha << "," << Beta <<")";
}



//xDivision:

xDivision::xDivision(xExpression *left, xExpression *right)
  : Left(left), Right(right)
{
}


xDivision::~xDivision()
{
  delete(Left);
  Left=0;
  delete(Right);
  Right=0;
}


void xDivision::AcceptVisitor(xPascalliVisitor &visitor)
{
  visitor.visit_xDivision((*this));
}


void xDivision::print()
{
  cout <<"(";
  if (Left) 
    {
  Left->print();
    }
  cout <<":";
  if (Right)
    {
      Right->print();
    }
  cout <<")"  <<" (" << Alpha << "," << Beta <<")";
}


//xAnd:

xAnd::xAnd(xCondition *left, xCondition *right)
  : Left(left), Right(right)
{
}


xAnd:: ~xAnd()
{
  delete(Left);
  Left=0;
  delete(Right);
  Right=0;
}


void xAnd::AcceptVisitor(xPascalliVisitor &visitor)
{
  visitor.visit_xAnd((*this));
}


void xAnd::print()
{
  cout <<"(";
  if (Left)
    {
  Left->print();
    }
  cout <<" AND ";
  if(Right)
    {
  Right->print();
    }
  cout <<")"  <<" (" << Alpha << "," << Beta <<")";
}


//xOr:

xOr::xOr(xCondition *left, xCondition *right)
  : Left(left), Right(right)
{
}


xOr:: ~xOr()
{
  delete(Left);
  Left=0;
  delete(Right);
  Right=0;
}


void xOr::AcceptVisitor(xPascalliVisitor &visitor)
{
  visitor.visit_xOr((*this));
}


void xOr::print()
{
  cout <<"(";
  if (Left)
    {
  Left->print();
    }
  cout <<" OR ";
  if(Right)
    {
  Right->print();
    }
  cout <<")"  <<" (" << Alpha << "," << Beta <<")";
}


//xNot:

xNot::xNot(xCondition *condition)
  : Expression(condition)
{
}


xNot:: ~xNot()
{
  delete(Expression);
  Expression=0;
}


void xNot::AcceptVisitor(xPascalliVisitor &visitor)
{
  visitor.visit_xNot((*this));
} 

void xNot::print()
{
  cout <<"NOT(";
  Expression->print();
  cout <<")"  <<" (" << Alpha << "," << Beta <<")";
}


//xGreater:

xGreater::xGreater(xExpression *left, xExpression *right)
  : Left(left), Right(right)
{
}


xGreater:: ~xGreater()
{
  delete(Left);
  Left=0;
  delete(Right);
  Right=0;
}


void xGreater::AcceptVisitor(xPascalliVisitor &visitor)
{
  visitor.visit_xGreater((*this));
} 


void xGreater::print()
{
  cout <<"(";
  if (Left)
    {
  Left->print();
    }
  cout <<" > ";
  if(Right)
    {
  Right->print();
    }
  cout <<")"  <<" (" << Alpha << "," << Beta <<")";  
}

//xGreaterEqual:

xGreaterEqual::xGreaterEqual(xExpression *left, xExpression *right)
  : Left(left), Right(right)
{
}


xGreaterEqual:: ~xGreaterEqual()
{
  delete(Left);
  Left=0;
  delete(Right);
  Right=0;
}


void xGreaterEqual::AcceptVisitor(xPascalliVisitor &visitor)
{
  visitor.visit_xGreaterEqual((*this));
}


void xGreaterEqual::print()
{
  cout <<"(";
  if (Left)
    {
  Left->print();
    }
  cout <<" >= ";
  if(Right)
    {
  Right->print();
    }
  cout <<")"  <<" (" << Alpha << "," << Beta <<")";  
}

//xEqual:

xEqual::xEqual(xExpression *left, xExpression *right)
  : Left(left), Right(right)
{
}


xEqual::~xEqual()
{
  delete(Left);
  Left=0;
  delete(Right);
  Right=0;
}


void xEqual::AcceptVisitor(xPascalliVisitor &visitor)
{
  visitor.visit_xEqual((*this));
} 


void xEqual::print()
{
  cout <<"(";
  if (Left)
    {
  Left->print();
    }
  cout <<" = ";
  if(Right)
    {
  Right->print();
    }
  cout <<")"  <<" (" << Alpha << "," << Beta <<")";  
}


//xSmaller:

xSmaller::xSmaller(xExpression *left, xExpression *right)
  : Left(left), Right(right)
{
}


xSmaller:: ~xSmaller()
{
  delete(Left);
  Left=0;
  delete(Right);
  Right=0;
}


void xSmaller::AcceptVisitor(xPascalliVisitor &visitor)
{
  visitor.visit_xSmaller((*this));
}


void xSmaller::print()
{
  cout <<"(";
  if (Left)
    {
  Left->print();
    }
  cout <<" < ";
  if(Right)
    {
  Right->print();
    }
  cout <<")"  <<" (" << Alpha << "," << Beta <<")";
}



//xSmallerEqual:

xSmallerEqual::xSmallerEqual(xExpression *left, xExpression *right)
  : Left(left), Right(right)
{
}


xSmallerEqual:: ~xSmallerEqual()
{
  delete(Left);
  Left=0;
  delete(Right);
  Right=0;
}


void xSmallerEqual::AcceptVisitor(xPascalliVisitor &visitor)
{
  visitor.visit_xSmallerEqual((*this));
}

void xSmallerEqual::print()
{
  cout <<"(";
  if (Left)
    {
  Left->print();
    }
  cout <<" <= ";
  if(Right)
    {
  Right->print();
    }
  cout <<")"  <<" (" << Alpha << "," << Beta <<")";
}



//xUnEqual:

xUnEqual::xUnEqual(xExpression *left, xExpression *right)
  : Left(left), Right(right)
{
}


xUnEqual:: ~xUnEqual()
{
  delete(Left);
  Left=0;
  delete(Right);
  Right=0;
}


void xUnEqual::print()
{
  cout <<"(";
  if (Left)
    {
  Left->print();
    }
  cout <<" <= ";
  if(Right)
    {
  Right->print();
    }
  cout <<")"  <<" (" << Alpha << "," << Beta <<")";
}


void xUnEqual::AcceptVisitor(xPascalliVisitor &visitor)
{
  visitor.visit_xUnEqual((*this));
}

















