/*
  Module for the purpose of compiling a parse tree
  Copyright (C) 1999  Philipp Cimiano Lavin

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

// ---------------------------------------------------------------------------- 
//  Description      : Module for the purpose of compiling a parse tree 
// ---------------------------------------------------------------------------- 
//  Remarks          : 
// ---------------------------------------------------------------------------- 
//  Modifications    : 
//    Achim     15.04.1999      created. 
//    Phil      19.05.1999      some minor changes.
//
// ---------------------------------------------------------------------------- 




#ifndef FRONTEND_H
#define FRONTEND_H




#include "base.h"
#include "xstring.h"
#include "commontype.h"




class xPascalliVisitor;
struct xNode;
struct xExpression;
struct xConstant;
struct xIdentifier;
struct xVariable;
struct xArray;
struct xFunctionCall;
struct xProgramNode;
struct xFunction;
struct xCommand;
struct xCommandBlock;
struct xAssignment;
struct xWhile;
struct xIfStatement;
struct xForLoop;
struct xAddition;
struct xSubtraction;
struct xMultiplication;
struct xDivision;
struct xCondition;
struct xBooleshExpression;
struct xAnd;
struct xOr;
struct xNot;
struct xGreaterEqual;
struct xGreater;
struct xEqual;
struct xSmaller;
struct xSmallerEqual;
struct xUnEqual;









struct xNode
{
  xString ResultAddress;

    xNode();
    virtual ~xNode();

  virtual void AcceptVisitor(xPascalliVisitor &visitor) = 0 ;
  virtual void print() = 0;
  TIndex Alpha;                                                // alpha is the number of the first command of a commandblock
  TIndex Beta;                                                 // betha is the number of the last command of a commandblock
};




struct xExpression : public xNode
{
};




struct xConstant : public xExpression
{
  xString Value;

    xConstant(xString value);
    virtual ~xConstant();

  virtual void AcceptVisitor(xPascalliVisitor & visitor);
  virtual void print();
};




struct xIdentifier : public xExpression
{
  TIndex   Number_Of_Arguments;
  xString  Name;
  EVarType Type;

    xIdentifier(xString name);
    virtual ~xIdentifier();

  virtual void AcceptVisitor(xPascalliVisitor &visitor);
  virtual void print();
  TBool operator==(xIdentifier identifier);

};


struct xVariable : public xIdentifier
{
    xVariable(xString name);
    virtual ~xVariable();
  
  virtual void AcceptVisitor(xPascalliVisitor &visitor);
  virtual void print(); 
};


struct xArray : public xIdentifier
{
  xExpression *Index;

    xArray(xString name, xExpression *index);
    virtual ~xArray();

  virtual void AcceptVisitor(xPascalliVisitor &visitor);
  virtual void print();
};



struct xFunctionCall : public xIdentifier
{
  xExpressionList &Arguments;

    xFunctionCall(xString name, xExpressionList &arguments);
    virtual ~xFunctionCall();

  virtual void AcceptVisitor(xPascalliVisitor & visitor);
  virtual void print();
};



struct xProgramNode : public xNode
{
  xFunctionList &Functions;

    xProgramNode(xFunctionList &functions);
    virtual ~xProgramNode();

  virtual void AcceptVisitor(xPascalliVisitor & visitor);
  virtual void print();
};


struct xFunction : public xNode
{
  xIdentifier Name;
  xIdentifierList &Arguments;
  xCommand *Command;

  xFunction(xIdentifier* name, xIdentifierList& arguments, xCommand* command);
  virtual ~xFunction();

  virtual void AcceptVisitor(xPascalliVisitor & visitor);
  virtual void print();
};


struct xCommand : public xNode
{
  virtual TBool variable_used(xIdentifier *identifier) = 0;
};




struct xCommandBlock : public xCommand
{
  xCommandBlock(xCommandList &commandlist);
  xCommandBlock();
  virtual ~xCommandBlock();

  xCommandList& CommandList;
  
  void addCommand(xCommand* newcommand);
  xCommand* getLastCommand();
  virtual void AcceptVisitor(xPascalliVisitor & visitor);
  virtual void print();
  virtual TBool variable_used(xIdentifier *identifier);
};


struct xAssignment : public xCommand
{
  xIdentifier *Destination;
  xExpression *Source;

    xAssignment(xIdentifier * destination, xExpression * source);
    virtual ~xAssignment();

  virtual void AcceptVisitor(xPascalliVisitor & visitor);
  virtual void print();
  virtual TBool variable_used(xIdentifier *identifier);
};




struct xWhile : public xCommand
{
  xCondition *Condition;
  xCommand *Command;

    xWhile(xCondition * condition, xCommand * command);
    virtual ~xWhile();

  virtual void AcceptVisitor(xPascalliVisitor & visitor);
  virtual void print();
  virtual TBool variable_used(xIdentifier *identifier);
};


struct xIfStatement : public xCommand
{
  xCondition *Condition;
  xCommand *IfTwig;
  xCommand *ElseTwig;		// This should be set to 0 (NULL) if no elestwig exists

    xIfStatement(xCondition *condition, xCommand *ifTwig, xCommand *elseTwig );
    virtual ~xIfStatement();

  virtual void AcceptVisitor(xPascalliVisitor & visitor);
  virtual void print();
  virtual TBool variable_used(xIdentifier *identifier);

};

struct xForLoop : public xCommand
{
  xIdentifier *Counter;
  xExpression *StartValue;
  xExpression *EndValue;
  xCommand *Command;

    xForLoop(xIdentifier *counter, xExpression *start, xExpression *end, xCommand *command);

    virtual ~xForLoop();

  virtual void AcceptVisitor(xPascalliVisitor & visitor);
  virtual void print();
  virtual TBool variable_used(xIdentifier *identifier);
};


struct xAddition : public xExpression
{
  xExpression *Left;
  xExpression *Right;

    xAddition(xExpression * left, xExpression * right);
    virtual ~xAddition();

  virtual void AcceptVisitor(xPascalliVisitor & visitor);
  virtual void print();
};




struct xSubtraction:public xExpression
{
  xExpression *Left;
  xExpression *Right;

    xSubtraction(xExpression * left, xExpression * right);
    virtual ~xSubtraction();

  virtual void AcceptVisitor(xPascalliVisitor & visitor);
  virtual void print();
};




struct xMultiplication : public xExpression
{
  xExpression *Left;
  xExpression *Right;

    xMultiplication(xExpression * left, xExpression * right);
    virtual ~xMultiplication();

  virtual void AcceptVisitor(xPascalliVisitor & visitor);
  virtual void print();
};




struct xDivision : public xExpression
{
  xExpression *Left;
  xExpression *Right;

    xDivision(xExpression * left, xExpression * right);
    virtual ~xDivision();

  virtual void AcceptVisitor(xPascalliVisitor & visitor);
  virtual void print();
};




struct xCondition : public xNode	//changed
{
};



struct xBooleshExpression : public xCondition	//changed
{
};




struct xAnd : public xBooleshExpression
{
  xCondition *Left;
  xCondition *Right;

    xAnd(xCondition *left, xCondition *right);
    virtual ~xAnd();

  virtual void AcceptVisitor(xPascalliVisitor & visitor);
  virtual void print();
};



struct xOr : public xBooleshExpression
{
  xCondition *Left;
  xCondition *Right;

    xOr(xCondition *left, xCondition *right);
    virtual ~xOr();

  virtual void AcceptVisitor(xPascalliVisitor & visitor);
  virtual void print();
};



struct xNot : xBooleshExpression
{
  xCondition *Expression;

    xNot(xCondition *condition);
    virtual ~xNot();

  virtual void AcceptVisitor(xPascalliVisitor & visitor);
  virtual void print();
};




struct xGreaterEqual : public xBooleshExpression
{
  xExpression *Left;
  xExpression *Right;

    xGreaterEqual(xExpression *left, xExpression *right);
    virtual ~xGreaterEqual();

  virtual void AcceptVisitor(xPascalliVisitor & visitor);
  virtual void print();
};




struct xGreater : public xBooleshExpression
{
  xExpression *Left;
  xExpression *Right;

    xGreater(xExpression *left, xExpression *right);
    virtual ~xGreater();

  virtual void AcceptVisitor(xPascalliVisitor & visitor);
  virtual void print();
};




struct xEqual : public xBooleshExpression
{
  xExpression *Left;
  xExpression *Right;

    xEqual(xExpression *left, xExpression *right);
    virtual ~xEqual();
  virtual void AcceptVisitor(xPascalliVisitor & visitor);
  virtual void print(); 
};




struct xSmaller : public xBooleshExpression
{
  xExpression *Left;
  xExpression *Right;

    xSmaller(xExpression *left, xExpression *right);
    virtual ~xSmaller();

  virtual void AcceptVisitor(xPascalliVisitor & visitor);
  virtual void print();
};




struct xSmallerEqual : public xBooleshExpression
{
  xExpression *Left;
  xExpression *Right;

    xSmallerEqual(xExpression *left, xExpression *right);
    virtual ~xSmallerEqual();

  virtual void AcceptVisitor(xPascalliVisitor & visitor);
  virtual void print();
};




struct xUnEqual : public xBooleshExpression
{
  xExpression *Left;
  xExpression *Right;

    xUnEqual(xExpression *left, xExpression *right);
    virtual ~xUnEqual();

  virtual void AcceptVisitor(xPascalliVisitor & visitor);
  virtual void print();
};




#endif









