/*
  plic compiler - first pass
  Copyright (C) 1999  Achim Kern

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

// ---------------------------------------------------------------------------- 
//  Description      : Implementation of the first pass of the Pascalli to
//                     Pascallino compiler
//
//                     This pass brings all functioncall in a form like
//                     a:=funccall(...);
//                     Also all variables are renamed to avoid conflicts
// ---------------------------------------------------------------------------- 
//  Creation         : 10.6.1999
// ---------------------------------------------------------------------------- 
//  Remarks          : One is a Variable containing the value 1.
//                     A corresponding line is added to the program in 
//                     xPascalliVisitor::visit_xProgramNode.
// ---------------------------------------------------------------------------- 
//  Modifications    : 28.7.2000
//                     rewrite of the function visit_xForloop
//                     This rewrite should close the bug #1,known as whiler-bug
//                     earned \      /  \      /  \      /
//                             \    /=)  \    /=)  \    /=)
//                              \--/      \--/      \--/
// ---------------------------------------------------------------------------- 




#include "plic_Backend1.h"
#include <iostream.h>



xPascalliVisitor_firstpass::xPascalliVisitor_firstpass()
{ 
  //  cout << "first pass started" << endl;
  LastVarNr = 0;
  InsertedCommands = 0;
};




xPascalliVisitor_firstpass::~xPascalliVisitor_firstpass()
{
};




void xPascalliVisitor_firstpass::
visit_xFunctionCall(xFunctionCall & node)
{
  xCommandBlock* Current = 0;
  xVariable*     TmpVar = 0;
  xAssignment*   TmpAssign = 0;
  
  for (TIndex Number = 0; Number < ~(node.Arguments); Number++)
  {
    node.Arguments[Number]->AcceptVisitor(*this);
    node.Arguments[Number] = (xExpression*)(Results.pop());
  };
  Current = InsertationPoint.pop();
  InsertationPoint.push(Current);
  TmpVar = MemManager.New();
  TmpAssign = new xAssignment(TmpVar, &node);
  node.Alpha += InsertedCommands;
  TmpAssign->Alpha = node.Alpha;
  Current->addCommand(TmpAssign);

  InsertedCommands++;

  Results.push(new xVariable(TmpVar->Name));
};



// translate a given for-loop into an equivalent while-loop:
// for i:= start to end do s; 
// 
// => c:=start
//    while c <= end do
//      s;
//      c := c +1;
//    wend
//
void xPascalliVisitor_firstpass::
visit_xForLoop(xForLoop & node)
{
    xCommandBlock* ResultBlock = new xCommandBlock;
    xCommandBlock* WhileBlock  = new xCommandBlock;
    xWhile*        WhileLoop   = 0;
    TIndex         LastCommand;

    // visit StartValue-Node and EndValue-Node
    node.EndValue->AcceptVisitor(*this);
    node.StartValue->AcceptVisitor(*this);

    // initialize the counter variable
    // Counter:= StartValue
    ResultBlock->addCommand(new xAssignment(new xIdentifier(node.Counter->Name), 
					    (xExpression*)(Results.pop())));
    ResultBlock->getLastCommand()->Alpha = node.Alpha + InsertedCommands;
    InsertedCommands++;

    // generate the new while node
    WhileLoop =   new xWhile(new xSmallerEqual(node.Counter, (xExpression*)(Results.pop())), 
			     WhileBlock);
    WhileLoop->Alpha = node.Alpha + InsertedCommands;
    
    // insert the while-loop to our result-block
    ResultBlock->addCommand(WhileLoop);

    // Traverse the commandblock of the former for-loop.
    // But first make sure that the results are written to the commandblock
    // of the new while-loop. Therefore first push WhileBlock to InsertationPoint
    // then do the translation.
    InsertationPoint.push(WhileBlock);
    node.Command->AcceptVisitor(*this);
    // add the result to WhileBlock
    WhileBlock->addCommand((xCommand*)(Results.pop()));
    // WhileBlock->getLastCommand()->Alpha = node.Alpha + InsertedCommands;
    // InsertedCommands++;
    LastCommand = WhileBlock->getLastCommand()->Alpha;
    // add the incrementation of the counter
    WhileBlock->addCommand(new xAssignment
			   (new xIdentifier(node.Counter->Name), 
			    new xAddition(new xIdentifier(node.Counter->Name), new xIdentifier("_One"))));
    WhileBlock->getLastCommand()->Alpha = LastCommand + 1;
    InsertedCommands++;
    
    // adjust the beta-value of the whilecommand
    WhileLoop->Beta = node.Beta + InsertedCommands;
    
    // clean the insertationPoint-stack
    InsertationPoint.pop();

    // push result
    Results.push(ResultBlock);
};





void xPascalliVisitor_firstpass::
visit_xAssignment(xAssignment & node)
{
  xPascalliTraverser::visit_xAssignment(node);
  node.Alpha += InsertedCommands;
};




void xPascalliVisitor_firstpass::
visit_xWhile(xWhile & node)
{
  xCommandBlock* NewLoopBody = 0;

  node.Condition->AcceptVisitor(*this);
  node.Alpha += InsertedCommands;
 
  NewLoopBody = new xCommandBlock;
  InsertationPoint.push(NewLoopBody);
  node.Command->AcceptVisitor(*this);
  InsertationPoint.pop();
  node.Beta += InsertedCommands;
  NewLoopBody->addCommand((xCommand*)(Results.pop()));

  node.Command = NewLoopBody;
  node.Condition = (xCondition*)(Results.pop());

  Results.push(&node);
};
