/*
  Implementation of the parse tree
  Copyright (C) 1999  Philipp Cimiano Lavin

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

// ---------------------------------------------------------------------------- 
//  Description      : Module for the purpose of compiling a parse tree 
// ---------------------------------------------------------------------------- 
//  Remarks          : 
//    Achim     31.01.1999 
// ---------------------------------------------------------------------------- 
//  Modifications    : 
//    Achim     31.01.1999      created. 
//    Achim     01.02.1999      after deep changes in design, 
//				redifinition of interface. 
//    Philipp   06.02.1999      after some minor changes in design, 
//			        redifinition of interface (xVisitor). 
//    Achim     09.02.1999      changes were made to xIdentifier, xArray
//                              xMemoryObject was added.
// ---------------------------------------------------------------------------- 




#ifndef COMPPANO_H
#define COMPPANO_H




#include "commontype.h"
#include "xstring.h"
#include <iostream.h>


class xParameter;
class xNode;
class xFunction;
class xCondition;
class xIdentifier;
class xVariable;
class xCommand;
class xPascallinoVisitor;
class xCommandBlock;
class xExpression;




struct xNode 
{ 
  xString ResultAddress;

  xNode();
  virtual ~xNode();

  virtual void AcceptVisitor(xPascallinoVisitor &visitor) = 0; 
  virtual void print()
    { cout << "XNODE is not printable!" << endl; }
}; 
 
 
 
 
struct xFunction : public xNode 
{ 
  xVariable   *Name; 
  xParameter  *ArgList; 
  xCommand    *Command; 
 
  xFunction(xVariable *name, xParameter *arglist, xCommand *command); 
  virtual ~xFunction();

  virtual void AcceptVisitor(xPascallinoVisitor &visitor); 
  virtual void print();
}; 
 
 

 
struct xParameter : public xNode 
{ 
  xArgList *ArgList; 
 
  xParameter(xArgList *arglist); 
  virtual ~xParameter();

  virtual void AcceptVisitor(xPascallinoVisitor &visitor); 
  virtual void print();
}; 
 
 

 
struct xCondition : public xNode
{ 
  xExpression   *Expression; 
 
  xCondition(xExpression *expression);
  virtual ~xCondition();

  virtual void AcceptVisitor(xPascallinoVisitor &visitor); 
  virtual void print();
}; 
 
 
 
 
struct xExpression : public xNode 
{
  virtual void print()
    { cout << "XEXPRESSION is not printable!" << endl; }
}; 
 
 
 
 
struct xCommand : public xNode 
{ 
  virtual void print()
    { cout << "XCOMMAND is not printable!" << endl;}
}; 
 
 
 
 
struct xCommandBlock : public xCommand 
{ 
  xCommandList  *CommandList; 
 
  xCommandBlock(xCommandList *commandlist); 
  virtual ~xCommandBlock();

  virtual void AcceptVisitor(xPascallinoVisitor &visitor); 
  virtual void print();
}; 
 
 
 

struct xAssignment : public xCommand 
{ 
  xIdentifier   *Destination; 
  xExpression   *Source; 
 
  xAssignment(xIdentifier *destination, xExpression *source); 
  virtual ~xAssignment();

  virtual void AcceptVisitor(xPascallinoVisitor &visitor); 
  virtual void print();
}; 
 
 
 
 
struct xWhile : public xCommand 
{ 
  xCondition            *Condition; 
  xCommand              *Command; 
 
  xWhile(xCondition *condition, xCommand *command); 
  virtual ~xWhile();

  virtual void AcceptVisitor(xPascallinoVisitor &visitor);
  virtual void print(); 
}; 
 
 
 
 
struct xSum : public xExpression 
{ 
  xExpression   *Left; 
  xExpression   *Right; 
 
  xSum(xExpression *left, xExpression *right); 
  virtual ~xSum();

  virtual void AcceptVisitor(xPascallinoVisitor &visitor);
  virtual void print();
}; 
 
 
 
 
struct xDiff : public xExpression 
{ 
  xExpression   *Left; 
  xExpression   *Right; 
 
  xDiff(xExpression *left, xExpression *right); 
  virtual ~xDiff();

  virtual void AcceptVisitor(xPascallinoVisitor &visitor); 
  virtual void print();
}; 
 
 
 
 
struct xConstant : public xExpression 
{ 
  const xString         Value; 
 
  xConstant(xString value);
  virtual ~xConstant();

  virtual void AcceptVisitor(xPascallinoVisitor &visitor); 
  virtual void print();
}; 




struct xMemoryObject
{
 public:
  TIndex* Address;
  xMemoryObject();
  ~xMemoryObject();
};
 
 
 
 
struct xIdentifier : public xExpression 
{ 
  xString        Name;
  xMemoryObject* Address;
  EVarType       Type;
  
  xIdentifier(xString name);
  virtual ~xIdentifier();
  virtual void print()
    { cout << "IDENTIFIER is not printable!" << endl; }
}; 
  



struct xVariable : public xIdentifier
{
  xVariable(xString name);
  virtual ~xVariable();

  virtual void AcceptVisitor(xPascallinoVisitor &visitor);
  virtual void print();
  TBool operator==(xVariable variable);
};



 
struct xArray : public xIdentifier 
{
  xExpression   *Index; 
 
  xArray(xString name, xExpression *index);
  virtual ~xArray();

  virtual void AcceptVisitor(xPascallinoVisitor &visitor); 
  virtual void print();
  TBool operator==(xArray array); 
}; 




typedef xList<TIndex> xRegList;




#endif



