/*
  ploc parser
  Copyright (C) 1999  Philipp Cimiano Lavin

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* Here follows the implementation of a Yacc-Parser for Pascallino.    */
/* The Parser includes and uses the module "visitor.h" in order to     */
/* create a Parse-Tree.                                                */
/* It also includes and uses "xset.h" in order to create a symbol-     */
/* table.							       */
/* The Parser uses two symbol-tables; one for the arrays and one for   */
/* the variables.						       */

%{
#include <iostream.h>
#include <time.h>
#include <stdlib.h>
#include "base.h"
#include "frontend.h"
#include "backend.h"
#include "commontype.h"
#include "xstring.h"
#include "xset.h"

int yylex();

void message_compilation_finished();

void message_abnormal_compilation();

int Status;

char* Outfile;

struct xVariableSet : public xSet<xVariable *> {
  inline TIndex operator+=(xVariable *object);
  inline void operator-=(xVariable *object);
  inline TIndex operator()(xVariable *object);
  inline TBool Contains(xVariable *object);
  inline TBool Contains(xVariable *object, TIndex &index);
};




TIndex xVariableSet::operator+=(xVariable *object)
{
  TIndex index;
  if (Contains(object, index)) return index;
  else return InternalInsert(Items, object);
}




void xVariableSet::operator-=(xVariable *object)
{
  TIndex index;
  if (Contains(object, index)) xList<xVariable *>::operator-=(index);
}




TIndex xVariableSet::operator()(xVariable *object)
{
  for (TIndex i = 0; i<Items; i++)
    if (*object == *List[i]) return i;
  throw xContainerException(ECCT_NOTFOUND);
}




TBool xVariableSet::Contains(xVariable *object)
{
  TIndex index;
  return Contains(object, index);
}




TBool xVariableSet::Contains(xVariable *object, TIndex &index)
{
  for (TIndex i = 0; i<Items; i++)
    if (*object == *List[i]) {
      index = i;
      return TRUE;
    }
  return FALSE;
}




xVariableSet Variable_Symboltable;;
xVariableSet Array_Symboltable;

xList<xNode *> Free_List;
xList<xMemoryObject *> Free_List_2;


xVariable *Function_Name;


TIndex Number_Functions=1;
TIndex Number_Parameters=0;
TIndex Number_Arrays=0;
TIndex Number_Variables=0;


void yyerror(char *s);

void yyerror2(char *s);

%}

/* The section until the first '%%' is called the Definition Section.  */
/* In this section the valid TOKENS returned by the Lexer are declared */
/* and assigned to a type.                                             */
/* Also the NON-TERMINALS are declared and assigned to a type.         */

%union
{
  char          *Key_word;
  char          *tIdentifier;
  char          *tConstant;
  xParameter    *tParameter;
  xCommandBlock *tCommandBlock;
  xCommand      *tCommand;
  xCondition    *tCondition;
  xExpression   *tExpression;
  xArray        *tArray;
}

%token <Key_word>    FUNCTION
%token <Key_word>    BEG
%token <Key_word>    END
%token <tIdentifier> IDENTIFIER
%token <tConstant>   CONSTANT
%token <Key_word>    WHILE
%token <Key_word>    DO

/* The tokens ';' ',' '#' ':' '=' '(' ')' don't need to be declared.  */

%type <tParameter>    arglist
%type <tCommandBlock> commandblock
%type <tCommand>      command
%type <tCondition>    condition
%type <tExpression>   expression
%type <tArray>        array
%%

/* Here begins the Rule Section.                                      */

program: FUNCTION IDENTIFIER 
         {
	   xVariable *identifier;
	   xString   symbol;


	   // a xString is created out of IDENTIFIER
	   symbol=xString($2);
	   
	   // a xVariable is created
	   identifier=new xVariable(symbol);
	   identifier->Type=Function;

	   // the global Function_Name is set!
	   Function_Name=identifier;

	 }

         '(' arglist ')' ';' command
	 {
	   TIndex      index;
	   xString     file;
	   xFunction   *function;
	   xVariable   *identifier;

	   identifier=Function_Name;

	   // checking if the identifier is already in the Symboltable
	   if (Variable_Symboltable.Contains(identifier))
	   // the identifier is already in the Symboltable
	   {
	     // the identifier is inserted into the symboltable in order
	     // to obtain the correct index
	     index=Variable_Symboltable+=identifier;

	     // the value of identifier is corrected
	     identifier=Variable_Symboltable[index];
	   }
	   else
	   // the identifier is not in the Symboltable
	   {
	     yyerror2("No return value!");
	     return 1;
	   }
	   
	   // cout << "function " << symbol <<" parsed!!!" <<endl;

	   // the xFunction to be compiled is created
	   function=new xFunction(identifier,$5,$8);

     	   // function is inserted into the Free_List
           Free_List+=function;

           // a xString is created out the Outfile
	   file=xString(Outfile);

	   // cout << "xvisitor is called with:" <<endl;
	   // cout << "Number of Parameters: " << Number_Parameters <<endl;
	   // cout << "Number of Arrays    : " << Number_Arrays     <<endl;
	   // cout << "Number of Variables : " << Number_Variables  <<endl;
	   // cout << "Function            : " << symbol            <<endl;
	   // cout << "File                : " << file              <<endl;

	   // the parse-tree is printed
	   // function->print();

	   // the xVisitor is created
	   xPascallinoVisitor compiler(Number_Parameters,
	   	  	               Number_Variables,
	    		               Number_Arrays,
	   		  	       function,
	   		 	       file);

           // all the nodes are destructed!!!

	   // cout << "Cleaning memory!!!\n";

	   for (TIndex index=0; index<~Free_List; index++)
	   {
	     delete(Free_List[index]);	
	   }

	   // cout << "Free_List OK!";

	   for (TIndex index=0; index<~Free_List_2; index++)
	   {
	     delete(Free_List_2[index]);
	   }

	   // cout << "Free_List_2 OK!";

	   // this rule raises only normal syntax-error!!!
	 }
       ;

commandblock: commandblock ';' command
	      {
		TIndex index;
		xCommandBlock *commandblock;
		xCommandList  *commandlist;

		commandlist=$1->CommandList;

		// the command is inserted into the commandlist
		index=(*commandlist)+=$3;

		// the commandblock to be returned is created
		commandblock=$1;

		// cout << "Commandblock parsed!\n";

		$$=commandblock;

		// this rule raises only normal syntax-error!!!
	      }
	    | command
	      {
		TIndex index;
		xCommandBlock *commandblock;
		xCommandList  *commandlist;

                // the commandlist is created
                commandlist=new xCommandList;

		// the command is inserted into the commandlist
		index=(*commandlist)+=$1;

		// a xCommandblock is created out of the commandlist
		commandblock=new xCommandBlock(commandlist);

		// commanblock is inserted into the Free_List
		Free_List+=commandblock;

		// the commandblock is returned
		$$=commandblock;

		// this rule raises only normal syntax-error!!!
	      }
	      ;

arglist: arglist ',' IDENTIFIER
	 {
	   TIndex      index;
	   xString     symbol;
	   xVariable   *variable;
	   xParameter  *parameter;
	   xArgList    *arglist;

	   // a xString is created out of identifier
	   symbol=xString($3);

	   // a xVariable is created
	   variable=new xVariable(symbol);

           Free_List+=variable;

	   // checking whether the parameter is already in the Symboltable
	   if (Variable_Symboltable.Contains(variable))
	   // the parameter is in the Symboltable!
	   {
	     yyerror("Formal parameter used before!!!");
	     return 1;;
	   }
	   else
	   // the parameter is not in the Symboltable!
	   {
	     // the number of parameters is increased
	     Number_Parameters++;

	     // a new xMemoryObject is created
	     variable->Address=new xMemoryObject();

	     Free_List_2+=variable->Address;

	     // cout << "Formal parameter " << symbol << " declared!\n";
	     // cout << "xVariable pointing to new MemoryObject created!\n";

	     // the parameter is inserted into the Symboltable
	     index=Variable_Symboltable+=variable;

	     // using already created ArgList
	     arglist=$1->ArgList;
	
	     // the parameter is inserted into the argList;
	     index=(*arglist)+=variable;

	     // the xParameter to be returned is created
	     parameter=$1;

	     // cout<< "Arglist parsed!\n"; 
	
	     $$=parameter;
	   }
	   // this rule raises an error if one of the parameters
	   // is already in the Symboltable!!!
	 }

       | IDENTIFIER
	 {
	   TIndex      index;
	   xString     symbol;
	   xVariable   *variable;
	   xParameter  *parameter;
	   xArgList    *arglist;

	   // a xString is created out of the IDENTIFIER
	   symbol=xString($1);

	   // a xVariable is created
	   variable=new xVariable(symbol);

           Free_List+=variable;

	   // checking whether the parameter is already in the Symboltable
	   if (Variable_Symboltable.Contains(variable))
	   // the parameter is in the Symboltable!
	   {
	     yyerror("Formal parameter used before");
	     return 1;
	   }
	   else
	   // the parameter is not in the Symboltable!
	   {
	     // the number of parameters is increased
	     Number_Parameters++;

	     variable->Address=new xMemoryObject();

	     Free_List_2+=variable->Address;

	     // cout << "Formal parameter " << symbol << " declared!\n";
	     // cout << "xVariable pointing to new MemoryObject created!\n";

	     // the parameter is inserted into the Symboltable
	     index=Variable_Symboltable+=variable;

	     // a xArgList is created
	     arglist=new xArgList;

	     // the parameter is inserted into the argList
	     index=(*arglist)+=variable;

	     // a xArglist is created to be returned
	     parameter=new xParameter(arglist);
		
	     Free_List+=parameter;

	     $$=parameter;
	   }

	   // this rule raises an error if one of the parameters
	   // is already in the Symboltable!!!
	 }
      |
 	{
	   xParameter  *parameter;
	   xArgList    *arglist;


	   // a xArgList is created
	   arglist=new xArgList;


	   // a xArglist is created to be returned
	   parameter=new xParameter(arglist);
		
	   Free_List+=parameter;

	   $$=parameter;		
	}	
        ;

command: BEG commandblock END
	   {
	     // cout << "BEGIN...END-Commandblock parsed!\n";

	     // the commandblock is returned
	      $$=$2;
	     // this rule raises normal syntax-error!!!
	   }
	 | WHILE condition DO command
	   {
	   xWhile *While;

	   // a xWhile is created and returned
	   While=new xWhile($2,$4);

	   Free_List+=While;

	   // cout << "While parsed!\n";

	   $$=While;

	     // this rule raises normal syntax-error!!!
	   }
	 | IDENTIFIER ':' '=' CONSTANT
	   {
	     TIndex index;
	     xString symbol;
	     xString constant;
	     xVariable *variable;
	     xConstant *source;
	     xAssignment *assignment;

	     // a xString is created out of the IDENTIFFIER
	     symbol=xString($1);

	     // a xVariable is created
	     variable=new xVariable(symbol);

             Free_List+=variable;

	     // checking whether the variable is in the Array_Symboltable
	     if (Array_Symboltable.Contains(variable))
	     // the variable is in the Array_Symboltable!
	     {
	       yyerror("Variable used as array before!!!");
	       return 1;
	     }
	     else
	     // the variable is not in the Array_Symboltable!
	     {
	       // checking if the variable is already in the Symboltable
	       if (Variable_Symboltable.Contains(variable))
	       // the variable is already in the Symboltable
	       {
		 // the variable is inserted into the Symboltable
		 // in order to obtain the correct index
		 index=Variable_Symboltable+=variable;

		 // the value of variable is corrected
		 variable=Variable_Symboltable[index];

	         // cout <<"Old variable " << symbol << " used!\n";
		 // cout <<"No new MemoryObject created!\n";

		 // a xString is created out of CONSTANT
		 constant=xString($4);

		 // a xConstant is created
		 source=new xConstant(constant);

 		 Free_List+=source;

		 // a xAssignment is created and returned
		 assignment=new xAssignment(variable,source);

		 // cout <<"Constant-assignment parsed!";

		 Free_List+=assignment;

		 $$=assignment;
	       }
	       else
	       // the variable is not in the Symboltable
	       {
		 // the number of variables is increased
		 Number_Variables++;

		 variable->Address=new xMemoryObject();

		 Free_List_2+=variable->Address;

		 // cout << "New variable " << symbol <<" declared!\n";
		 // cout << "xVariable pointing to new MemoryObject created!\n";

		 // the variable is inserted into the Symboltable
		 index=Variable_Symboltable+=variable;

		 // a xString is created out of CONSTANT
		 constant=xString($4);

		 // a xConstant is created
		 source=new xConstant(constant);

 		 Free_List+=source;

		 // a xAssignment is created and returned
		 assignment=new xAssignment(variable,source);

		 // cout <<"Constant-assignment parsed!\n";

		 Free_List+=assignment;

		 $$=assignment;
	       }
	     }

	     // this rule raises an error if the variable has been used
	     // as an array before!!!
	   }
	 | IDENTIFIER ':' '=' expression
	   {
	     TIndex      index;
	     xString     symbol;
	     xVariable   *variable;
	     xAssignment *assignment;

	     // a xString is created out of IDENTIFIER
	     symbol=xString($1);

	     // a xVariable is created
	     variable=new xVariable(symbol);

	     Free_List+=variable;

	     // checking whether the variable is in the Array_Symboltable
	     if (Array_Symboltable.Contains(variable))
	     // the variable is in the Array_Symboltable!
	     {
	       yyerror("Variable used as array before!!!");
	       return 1;
	     }
	     else
	     // the variable is not in the Array_Symboltable!
	     {
	       // checking if the variable is in the Symboltable
	       if (Variable_Symboltable.Contains(variable))
	       // the variable is in the Symboltable
	       {
		 // the variable is inserted into the Symboltable
		 // in order to obtain the correct index
		 index=Variable_Symboltable+=variable;

		 // the value of variable is corrected
		 variable=Variable_Symboltable[index];

		 // cout <<"Old variable " << symbol << " used!\n";
		 // cout <<"No new MemoryObject created!\n";

		 // a xAssignment is created and returned
		 assignment=new xAssignment(variable,$4);

		 Free_List+=assignment;

	 	 // cout << "Expression-assginment parsed!\n";

		 $$=assignment;
	       }
	       else
	       // the variable is not in the Symboltable
	       {
		 // the number of variables is increased
		 Number_Variables++;

		 variable->Address=new xMemoryObject;

		 Free_List_2+=variable->Address;

		 // cout <<"New variable " << symbol << " declared!\n";
		 // cout <<"xVariable pointing to new MemoryObject created!\n";

		 // the variable is inserted into the Symboltable
		 index=Variable_Symboltable+=variable;

		 // a xAssignment is created and returned
		 assignment=new xAssignment(variable,$4);

		 Free_List+=assignment;

		 // cout << "Expression-assignment parsed!\n";

		 $$=assignment;
	       }
	     }

	     // this rule raises an error if the variable has been used
	     // as an array before!!!
	   }
	 | IDENTIFIER ':' '=' array
	   {
	     TIndex      index;
	     xString     symbol;
	     xVariable   *variable;
	     xAssignment *assignment;

	     // a xString is created out of IDENTIFIER
	     symbol=xString($1);

	     // a xVariable is created
	     variable=new xVariable(symbol);

	     Free_List+=variable;

	     // checking whether the variable is in the Array_Symboltable
	     if (Array_Symboltable.Contains(variable))
	     // the variable is in the Array_Symboltable!
	     {
	       yyerror("Variable used as array before!!!");
	       return 1; 
	     }
	     else
	     // the variable is not in the Array_Symboltable!
	     {
	       // checking if the variable is already in the Symboltable
	       if (Variable_Symboltable.Contains(variable))
	       // the variable is already in the Symboltable
	       {
		 // the variable is inserted into the Symboltable
		 // in order to obtain the correct index
		 index=Variable_Symboltable+=variable;

		 // the value of variable is corrected
		 variable=Variable_Symboltable[index];

		 // cout <<"Old variable " << symbol << " used!\n";
		 // cout <<"No new MemoryObject created!\n";

		 // a xAssignment is created and returned
		 assignment=new xAssignment(variable,$4);

		 Free_List+=assignment;

		 $$=assignment;
	       }
	       else
	       // the variable is not in the Symboltable
	       {
		 // the number of variables is increased
		 Number_Variables++;

		 variable->Address=new xMemoryObject();
		 
		 Free_List_2+=variable->Address;

		 // cout <<"New variable " << symbol << " declared!\n";
		 // cout <<"xVariable pointing to new MemoryObject created!\n"; 

		 // the variable is inserted into the Symboltable
		 index=Variable_Symboltable+=variable;

		 // a xAssignment is created and returned
		 assignment=new xAssignment(variable,$4);

		 Free_List+=assignment;

		 // cout <<"Array_assignment parsed!\n";

		 $$=assignment;
	       }
	     }

	     // this rule raises an error if the variable has been used
	     // as an array before!!!
	   }
	 | array ':' '=' IDENTIFIER
	   {
	     TIndex      index;
	     xString     symbol;
	     xVariable   *variable;
	     xAssignment *assignment;

	     // a xString is created out of IDENTIFIER
	     symbol=xString($4);

	     // a xVariable is created
	     variable=new xVariable(symbol);

	     Free_List+=variable;

	     // checking if one of the identifiers is the function-identifier
	     if ((*variable)==(*Function_Name))
	       {
		 yyerror("function-identifier used at right side of an assignment!!!");
		 return 1;
	       }
	     


	     // checking whether the variable is in the Array_Symboltable
	     if (Array_Symboltable.Contains(variable))
	     // the variable is in the Array_Symboltable
	     {
	       yyerror("Variable used as array before!!!");
	       return 1;
	     }
	     else
	     // the variable is not in the Array_Symboltable
	     {              
	       // checking if the variable is in the Symboltable
	       if (Variable_Symboltable.Contains(variable))
	       // the variable is in the Symboltable
	       {
		 // the variable is inserted into the Symboltable
		 // in order to obtain the correct index
		 index=Variable_Symboltable+=variable;

		 // the value of variable is corrected
		 variable=Variable_Symboltable[index];

                 // checking if the variable is the function-identifier
		 if (variable->Type==Function)
		 // the  variable is the function-identifier
		 {
		   yyerror("function-identifier used as an assignable expression!!!");
		   return 1;
		 }

		 // cout <<"Old variable " << symbol << " used!\n";
		 // cout <<"No new MemoryObject created!\n";

		 // the xAssignment is created and returned
		 assignment=new xAssignment($1,variable);

		 Free_List+=assignment;

		 // cout <<"Array-Identifier-assignment parsed!\n";

		 $$=assignment;
	       }
	       else
	       // the variable is not in the Symboltable
	       {
		 // the number of variables is increased
		 Number_Variables++;

		 variable->Address=new xMemoryObject();
		 
		 Free_List_2+=variable->Address;

		 // cout <<"New variable " << symbol << " declared!\n";
		 // cout <<"xVariable pointing to new MemoryObject created!\n"; 	

		 // the variable is inserted into the Symboltable
		 index=Variable_Symboltable+=variable;

		 // the xAssignment is created and returned
		 assignment=new xAssignment($1,variable);

		 Free_List+=assignment;

		 // cout <<"Array-Identifier-assignment parsed!\n";

		 $$=assignment;
	       }
	     }

	     // this rule raises an error if the variable has been used
	     // as an array before!!!
	   }
	 ;

condition: '(' IDENTIFIER '#' CONSTANT  ')'
	   {
	     TIndex      index;
	     xString     symbol;
	     xVariable   *variable;
	     xCondition  *condition;

	     // a xString is created out of IDENTIFIER
	     symbol=xString($2);

	     // a xVariable is created
	     variable=new xVariable(symbol);

	     Free_List+=variable;
		

	     // checking whether the variable is in the Array_Symboltable
	     if (Array_Symboltable.Contains(variable))
	     // the variable is in the Array_Symboltable
	     {
	       yyerror("Variable used as array before!!!");
	       return 1;
	     }
	     else
	     // the variable is not in the Array_Symboltable
	     {
	       // checking if the variable is in the Symboltable
	       if (Variable_Symboltable.Contains(variable))
	       // the variable is already in the Symboltable
	       {
		 // the variable is inserted into the Symboltable
		 // in order to obtain the correct index
		 index=Variable_Symboltable+=variable;

		 // the value of variable is corrected
		 variable=Variable_Symboltable[index];

		 // cout <<"Old variable " << symbol << " used!\n";
		 // cout <<"No new MemoryObject created!\n"; 

		 // a xCondition is created and returned
		 condition=new xCondition(variable);
		
		 Free_List+=condition;

		 // cout << "Condition parsed!\n";

		 $$=condition;
	       }
	       else
	       // the variable is not in the Symboltable
	       {
		 // the number of variables is increased
		 Number_Variables++;

		 variable->Address=new xMemoryObject;

		 Free_List_2+=variable->Address;

		 // cout <<"New variable " << symbol << " declared!\n";
		 // cout <<"xVariable pointing to new MemoryObject created!\n";

		 // the variable is inserted into the Symboltable
		 index=Variable_Symboltable+=variable;

		 // a xCondition is created and returned
		 condition=new xCondition(variable);

		 Free_List+=condition;

		 // cout << "Condition parsed!\n";

		 $$=condition;
	       }
	     }

	     // this rule raises an error if the variable has been used
	     // as an array before!!!
	   }
	 ;

expression:
	    '(' IDENTIFIER '+' IDENTIFIER ')'
	    {
	      TIndex    index;
	      xString   symbol1;
	      xString   symbol2;
	      xVariable *variable1;
	      xVariable *variable2;
	      xSum      *sum;

	      // the xString's are created
	      symbol1=xString($2);
	      symbol2=xString($4);

	      // the xVariable's are created
	      variable1=new xVariable(symbol1);

	      Free_List+=variable1;

	      variable2=new xVariable(symbol2);

	      Free_List+=variable2;

	      // checking if one of the identifiers is the function-identifier
	      if (((*variable1)==(*Function_Name)) || ((*variable2)==(*Function_Name)))
		{
		  yyerror("function-identifier used at right side of an assignment!");
		  return 1;
		}


	      // checking if one of the variables is in the Array_Symboltable
	      if ( (Array_Symboltable.Contains(variable1))
		 ||(Array_Symboltable.Contains(variable2)))
	      // one of them is in the Array_Symboltable
	      {
		yyerror("Variable used as an array before!!!");
	 	return 1;
	      }
	      else
	      // none of them is in the Array_Symboltable
	      {
		// checking if variable1 is in the Symboltable
		if (Variable_Symboltable.Contains(variable1))
		// the variable1 is in the Symboltable
		{
		  // the variable1 is inserted into the Symboltable
		  // in order to obain the correct index
		  index=Variable_Symboltable+=variable1;

		  // the value of the variable1 is corrected
		  variable1=Variable_Symboltable[index];

		  // cout <<"Old variable " << symbol1 << " used!\n";
		  // cout <<"No new MemoryObject created!\n"; 
		}
		else
		// the variable1 is not in the Symboltable
		{
		  // the number of variables is increased
		  Number_Variables++;

		  variable1->Address=new xMemoryObject;

		  Free_List_2+=variable1->Address;

		  // cout <<"New variable " << symbol1 << " declared!\n";
		  // cout <<"xVariable pointing to new MemoryObject created!\n";

		  // the variable1 is inserted into the Symboltable
		  index=Variable_Symboltable+=variable1;
		}

		// checking if variable2 is in the Symboltable
		if (Variable_Symboltable.Contains(variable2))
		// the variable2 is in the Symboltable
		{
		  // the variable2 is inserted into the Symboltable
		  // in order to obain the correct index
		  index=Variable_Symboltable+=variable2;

		  // the value of the variable1 is corrected
		  variable2=Variable_Symboltable[index];

		  // cout <<"Old variable " << symbol2 << " used!\n";
		  // cout <<"No new MemoryObject created!\n"; 
		}
		else
		// the variable2 is not in the Symboltable
		{
		  // the number of variables is increased
		  Number_Variables++;

		  variable2->Address=new xMemoryObject();

		  Free_List_2+=variable2->Address;

		  // cout <<"New variable " << symbol2 << " declared!\n";
		  // cout <<"xVariable pointing to new MemoryObject created!\n";

		  // the variable2 is inserted into the Symboltable
		  index=Variable_Symboltable+=variable2;
		}
		// a xSum is created and returned
		sum=new xSum(variable1,variable2);

		Free_List+=sum;

		// cout <<"+expression parsed!\n";

		$$=sum;
	      }

	      // this rule raises an error if one of the two variables
	      // has been used as an array before!!!
	    }
	  | '(' IDENTIFIER '-' IDENTIFIER ')'
	    {
	      TIndex    index;
	      xString   symbol1;
	      xString   symbol2;
	      xVariable *variable1;
	      xVariable *variable2;
	      xDiff     *diff;

	      // the xString's are created
	      symbol1=xString($2);
	      symbol2=xString($4);

	      // the xVariable's are created
	      variable1=new xVariable(symbol1);

	      Free_List+=variable1;

	      variable2=new xVariable(symbol2);

	      Free_List+=variable2;

	      
	      // checking if one of the identifiers is the function-identifier
	      if (((*variable1)==(*Function_Name)) || ((*variable2)==(*Function_Name)))
		{
		  yyerror("function-identifier used at right side of an assignment!");
		  return 1;
		}

	      // checking if one of the variables is in the Array_Symboltable
	      if ( (Array_Symboltable.Contains(variable1))
		 ||(Array_Symboltable.Contains(variable2)))
	      // one of them is in the Array_Symboltable
	      {
		yyerror("Variable used as an array before!!!");
		return 1;
	      }
	      else
	      // none of them is in the Array_Symboltable
	      {
		// checking if variable1 is in the Symboltable
		if (Variable_Symboltable.Contains(variable1))
		// the variable1 is in the Symboltable
		{
		  // the variable1 is inserted into the Symboltable
		  // in order to obain the correct index
		  index=Variable_Symboltable+=variable1;

		  // the value of the variable1 is corrected
		  variable1=Variable_Symboltable[index];

		  // cout <<"Old variable " << symbol1 << " used!\n";
 		  // cout <<"No new MemoryObject created!\n";
		}
		else
		// the variable1 is not in the Symboltable
		{
		  // the number of variables is increased
		  Number_Variables++;

		  variable1->Address=new xMemoryObject();

		  Free_List_2+=variable1->Address;

		  // cout <<"New variable " << symbol1 << " declared!\n";
		  // cout <<"xVariable pointing to new MemoryObject created!\n"; 

		  // the variable1 is inserted into the Symboltable
		  index=Variable_Symboltable+=variable1;
		}

		// checking if variable2 is in the Symboltable
		if (Variable_Symboltable.Contains(variable2))
		// the variable2 is in the Symboltable
		{
		  // the variable2 is inserted into the Symboltable
		  // in order to obain the correct index
		  index=Variable_Symboltable+=variable2;

		  // the value of the variable1 is corrected
		  variable2=Variable_Symboltable[index];

		  // cout <<"Old variable " << symbol2 << " used!\n";
		  // cout <<"No new MemoryObject created!\n";
		}
		else
		// the variable2 is not in the Symboltable
		{
		  // the number of variables is increased
		  Number_Variables++;

		  variable2->Address=new xMemoryObject();

		  Free_List_2+=variable2->Address;

		  // cout <<"New variable " << symbol2 << " declared!\n";
		  // cout <<"xVariable pointing to new MemoryObject created!\n"; 

		  // the variable2 is inserted into the Symboltable
		  index=Variable_Symboltable+=variable2;
		}
		// a xSum is created and returned
		diff=new xDiff(variable1,variable2);

		Free_List+=diff;

		// cout <<"-expression parsed!\n";

		$$=diff;

	      }

	      // this rule raises an error if one of the two variables
	      // has been used as an array before!!
	    }
          | IDENTIFIER
            {
	      xVariable  *identifier;
	      xString    symbol;
	      TIndex     index;
	      
	      // a xString is created out of IDENTIFIER
	      symbol= xString($1);

	      // a xVariable is created out of IDENTIFIER
	      identifier=new xVariable(symbol);
	      
	      //checking if the identifieris in the Array_Symboltable
	      if (Array_Symboltable.Contains(identifier))
		{
		  yyerror("Variable used as an array before!!!");
		}
	      
	      // checking ih the identifier is the function_identifier
	      if ((*identifier)==(*Function_Name))
		{
		  yyerror("function-identifier used at right side of an assignment!");
		  return 1;
		}

	      // the identifier is inserted into the Free_List
	      Free_List+=identifier;
	      
	      // checking if the identifier is in the Symboltable
	      if (Variable_Symboltable.Contains(identifier))
		// the identifier is in the Symboltable
		{
		  // the identifier is inserted into the Symboltable
		  // in order to obain the correct index
		  index=Variable_Symboltable+=identifier;

		  // the value of the identifier is corrected
		  identifier=Variable_Symboltable[index];
		  
		  // cout <<"Old variable " << symbol1 << " used!\n";
		  // cout <<"No new MemoryObject created!\n"; 
		}
	      else
		// the variable1 is not in the Symboltable
		{
		  // the number of variables is increased
		  Number_Variables++;

		  identifier->Address=new xMemoryObject;
		  
		  Free_List_2+=identifier->Address;
		  
		  // cout <<"New variable " << symbol1 << " declared!\n";
		  // cout <<"xVariable pointing to new MemoryObject created!\n";
		  // the variable1 is inserted into the Symboltable
		  index=Variable_Symboltable+=identifier;
		}

	      $$=identifier;

	      // this rule raises an error if the identifier has been
	      // used as an array before and if the identifier is
	      // the function-identifier
	    }
	  ;

array:
       IDENTIFIER '[' IDENTIFIER ']'
       {
	 TIndex    index;
	 xString   array_symbol;
	 xString   index_symbol;
	 xVariable *array_identifier;
	 xVariable *array_index;
	 xArray    *array;

	 // xString's are created out of the IDENTIFIERs
	 array_symbol=xString($1);
	 index_symbol=xString($3);

	 // xVariable's are created
	 array_index=     new xVariable(index_symbol);

	 Free_List+=array_index;

	 array_identifier=new xVariable(array_symbol);

	 Free_List+=array_identifier;

	 // checking whether the array_identifier is in the Symboltable
	 if (Variable_Symboltable.Contains(array_identifier))
	 // the array_identifier is in the Variable_Symboltable
	 {
	   yyerror("Array used as variable before!\n");
	   return 1;
	 }
	 else
	 // the array_identifier is not in the Variable_Symboltable
	 {
	   // checking whether the array_index is in the Array_Symboltable
	   if (Array_Symboltable.Contains(array_index))
	   // the array_index is in the Array_Symboltable
	   {
	     yyerror("Variable in index used as array before!\n");
	     return 1;
	   }
	   else
	   // the array_index is not in the Array_Symboltable
	   {
	     // checking if the array_index is already in the Symboltable
	     if (Variable_Symboltable.Contains(array_index))
	     // the array_index is already in the Symboltable
	     {
	       // the array_index is inserted into the Symboltable
	       // in order to obtain the correct index
	       index=Variable_Symboltable+=array_index;

	       // the value of array_index is corrected
	       array_index=Variable_Symboltable[index];

	       // cout <<"Old variable " << index_symbol << " used!\n";
	       // cout <<"No new MemoryObject created!\n"; 
	     }
	     else
	     // the array_index is not in the Symboltable
	     {
	       // the number of variables is increased
	       Number_Variables++;

	       array_index->Address=new xMemoryObject();

	       Free_List_2+=array_index->Address;

	       // cout <<"New variable " << index_symbol << " declared!\n";
	       // cout <<"xVariable pointing to new MemoryObject created!\n";

	       // the array_index is inserted into the Symboltable
	       index=Variable_Symboltable+=array_index;
	     }

	     // checking if the array_identifier is already in the
	     // Array_Symboltable
	     if (Array_Symboltable.Contains(array_identifier))
	     // the array_identifier is in the Array_Symboltable
	     {

		// now we insert it into the Array_Symboltable in order
                // to obtain the correct index 
		index=Array_Symboltable+=array_identifier;

 	        // the arrays have to point to the same MemoryObject	
		array_identifier->Address=(Array_Symboltable[index])->Address;

		// cout <<"Old array " << array_symbol << " used!\n";
	        // cout <<"No new MemoryObject created!\n";
		// cout <<"xArray pointing to OLD MemoryObject created!\n"; 	
	     }
	     else
	     // the array_identifier is not in the Array_Symboltable
	     {
	       // the number of arrays is increased
	       Number_Arrays++;

	       array_identifier->Address=new xMemoryObject();	

	       Free_List_2+=array_identifier->Address;

	       // cout <<"New array " << array_symbol << " declared!\n";
	       // cout <<"xArray pointing to NEW MemoryObject created!\n";
	       // cout <<"Address: " << array_identifier->Address;

	       // the array_identifier is inserted into the
	       // Array_Symboltable
	       index=Array_Symboltable+=array_identifier;
	     }
	     // the xArray is created to be returned
	     array=new xArray((array_identifier->Name),array_index);
	     array->Address=array_identifier->Address;

	     Free_List+=array;

	     $$=array;
	   }
	 }

	 // this rule raises an error if the array_identifier has
	 // been used as a variable before and if the array_index
	 // has been used as an array before!!!
       }
     ;

%%

extern FILE *yyin;
FILE *file;
int main(int argc, char **argv)
{
  try {
    if (argc==3)
    {
      Outfile=argv[2];
      file=fopen(argv[1], "r");
      if (file)
      {
	ungetc('\n',file);
        yyin=file;
      }		
      else
      {
	cout <<"Could not open file: " << argv[1] <<"\n";
	return 1; 
      }
	  
	}
    else
    {
      cout <<"syntax: ploc <infile> <outfile>!\n";
      return 1;	
    }
		
    do
      {
	Status=yyparse();
      }
    while(!feof(yyin));

    if (Status==1)
      {
	message_abnormal_compilation();
	return 1;
      }
    else 
      {
	message_compilation_finished();
	return 0;
      }

  }catch (xBaseException &x) {
    char msg[10000]; // fixme design error in xbaseexception
    x.GetDescription(msg);
    cout << msg << "\n\n\n";
    message_abnormal_compilation();
    return 1;
  }catch (...) {
    cout << "unknnown exception caught\n\n\n\n";
    message_abnormal_compilation();
    return 1;
  };
  
};
  


void message_abnormal_compilation()
{
  struct tm *pTime;
  time_t t;
  
  t=time(NULL);
  pTime=localtime(&t);
  
  cout << "Compilation exited abnormally at " << asctime(pTime);
}

void message_compilation_finished()
{
struct tm *pTime;
time_t t;

t=time(NULL);
pTime=localtime(&t);

 cout << "Compilation finished at " << asctime(pTime);
}



