/*
  theosim main widget
  Copyright (C) 1999  Erik Thiele

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "mw.hpp"
#include "mw.moc"

#include <qapp.h>
#include <qlayout.h>
#include <qmessagebox.h>
#include <qfiledialog.h>
#include <qpainter.h>
#include <qpushbutton.h>
#include <qtextstream.h>
#include <qregexp.h>
#include <unistd.h>
#include <errno.h>
#include <stdlib.h>
#include <signal.h>
#include <fcntl.h>

#include "config.h"
#include "theosim.xpm"

void signal_describe(int sig, QString &s)
{
  switch (sig) {
  case 1: s="SIGHUP"; return;
  case 2: s="SIGINT"; return;
  case 3: s="SIGQUIT"; return;
  case 4: s="SIGILL"; return;
  case 6: s="SIGABRT"; return;
  case 8: s="SIGFPE"; return;
  case 9: s="SIGKILL"; return;
  case 11: s="SIGSEGV"; return;
  case 13: s="SIGPIPE"; return;
  case 14: s="SIGALRM"; return;
  case 15: s="SIGTERM"; return;
  default: s.sprintf("SIGNAL %d",sig); return;
  };
};

external::~external()
{
  ::close(output[0]);
  ::close(error[0]);
  terminatechild();
};

external::external (const char *messagestring, 
		    const char *filename, char *const *argvlist, QWidget *parent,const char *name)
  : QDialog (parent, name, TRUE)
{
  result=interrupted;

  resize (0, 0);

  QVBoxLayout *v = new QVBoxLayout (this, 5);

  message = new QLabel (this);
  message->setText(messagestring);
  message->setAlignment(AlignCenter);
  message->setMinimumHeight (20);
  v->addWidget (message, 0);

  exbox = new QMultiLineEdit (this);
  QFont f( "fixed", 10 );
  f.setFixedPitch(TRUE);
  exbox->setFont(f);
  exbox->setMinimumSize (600, 300);
  v->addWidget (exbox, 1);

  status = new QLabel (this);
  status->setText("Running...");
  status->setAlignment(AlignVCenter);
  status->setMinimumHeight (20);
  v->addWidget (status, 0);

  QHBoxLayout *h = new QHBoxLayout (5);
  v->addLayout (h, 0);
  h->addStretch (1);
  ok = new QPushButton ("Ok", this);
  connect(ok,SIGNAL(clicked()),this,SLOT(okbut()));
  ok->setEnabled(FALSE);
  ok->setFixedSize (ok->sizeHint());
  h->addWidget (ok, 0);
  h->addStretch (1);
  interrupt = new QPushButton ("Interrupt", this);
  connect(interrupt,SIGNAL(clicked()),this,SLOT(interruptbut()));
  interrupt->setFixedSize (interrupt->sizeHint());
  h->addWidget (interrupt, 0);
  h->addStretch (1);

  v->activate ();

  ::pipe(output);
  ::pipe(error);

  outputnotify = new QSocketNotifier (output[0],QSocketNotifier::Read,this);
  connect(outputnotify,SIGNAL(activated(int)),this,SLOT(activated(int)));
  errornotify = new QSocketNotifier (error[0],QSocketNotifier::Read,this);
  connect(errornotify,SIGNAL(activated(int)),this,SLOT(activated(int)));
  
  if (!(childpid=::fork())) { // child
    ::close(output[0]);
    ::dup2(output[1],1);
    ::close(output[1]);

    ::close(error[0]);
    ::dup2(error[1],2);
    ::close(error[1]);

    ::close(0);
    ASSERT(::open("/dev/null",O_RDONLY)==0);

    ::execvp(filename,argvlist);
    fprintf(stderr,"Execution failed.\n%s\n",strerror(errno));

    ::_exit(1);
  }else{ // mother
    ::close(output[1]);
    ::close(error[1]);
  };
};

void external::terminatechild()
{
  ::kill(childpid,SIGINT);
  ::waitpid(-1,&childtermstatus,0);
  ok->setEnabled(TRUE);
  interrupt->setEnabled(FALSE);
  QString s;
  if (WIFEXITED(childtermstatus)) {
    if (WEXITSTATUS(childtermstatus)==0) {
      result = success;
      s="Success.";
    }else{
      result = failure;
      s="Error !";
    };
  }else if (WIFSIGNALED(childtermstatus)) {
    if (WTERMSIG(childtermstatus)==SIGINT) {
      result = interrupted;
      s="Interrupted";
    }else{
      result = failure;
      QString sigdesc;
      signal_describe(WTERMSIG(childtermstatus),sigdesc);
      s="Terminated by signal: ";
      s+=sigdesc;
      if (WCOREDUMP(childtermstatus))
	s+=" (core dumped)";
    };
  }else{
    result=failure;
    s="unknown way of termination";
  };
  status->setText(s);
};

void external::activated(int hand)
{
  int retval;
  QString msg ((hand==output[0]) ? "> " : "! ");
  char c;

 again:
  retval=read(hand,&c,1);
  if (retval==1) {
    if (c!='\n') {
      msg+=c;
      goto again;
    };
  }else{
    if (hand==output[0]) {
      outputnotify->setEnabled(FALSE);
    }else{
      errornotify->setEnabled(FALSE);
    };
    if ((!outputnotify->isEnabled())&&(!errornotify->isEnabled())) terminatechild();
  };

  if (msg.length()>2) exbox->insertLine(msg);
};

void external::okbut()
{
  accept();
};

void external::interruptbut()
{
  terminatechild();
};


bgwdg::bgwdg (QWidget *parent, const char *name)
  : QWidget (parent,name)
{
  setMinimumSize(pixmap_theosim->width()+10,pixmap_theosim->height()+10);
};

void bgwdg::paintEvent ( QPaintEvent * )
{
  QPainter p(this);
  p.drawPixmap((width()-pixmap_theosim->width())/2,
	       (height()-pixmap_theosim->height())/2,*pixmap_theosim);
};


buffer::buffer (QWidget *parent, const char *name)
  : QWidget (parent,name)
{
  changed=false;
  filetype=RAM;

  QVBoxLayout *v = new QVBoxLayout (this,5);

  editor = new QMultiLineEdit(this);
  QFont f( "fixed", 12 );
  f.setFixedPitch(TRUE);
  editor->setFont(f);
  connect(editor,SIGNAL(textChanged()),this,SLOT(textChanged()));
  v->addWidget(editor,1);

  status = new QLabel (this);
  status->setMinimumSize(400,20);
  v->addWidget(status,0);

  v->activate();

  updatestatus();
};

void buffer::updatestatus()
{
  QString s;
  switch (filetype) {
  case PASCALLI: s+="PASCALLI"; break;
  case PASCALLINO: s+="PASCALLINO"; break;
  case RAM: s+="RAM"; break;
  };
  s+="    ";
  s+=changed ? "changed" : "not changed";
  s+="    ";
  s+=filename;
  status->setText(s);
};

void buffer::textChanged()
{
  if (!changed) {
    changed = true;
    updatestatus();
  };
};


mainw::~mainw()
{
  delete pixmap_theosim;
};

mainw::mainw(QWidget *parent,const char *name)
  : QWidget (parent,name)
{
  resize(0,0);

  pixmap_theosim = new QPixmap ((const char **)theosim_xpm);

  QVBoxLayout *v = new QVBoxLayout(this,5);

  menubar = new QMenuBar( this );

  filepop = new QPopupMenu;
  filepop->insertItem("&Open",this,SLOT(fileopen()));
  id_file_save=filepop->insertItem("&Save",this,SLOT(filesave()),0,-2);
  id_file_close=filepop->insertItem("&Close",this,SLOT(fileclose()),0,-2);
  filepop->insertSeparator();
  filepop->insertItem( "&Quit", this, SLOT(filequit()) );
  menubar->insertItem( "&File", filepop );

  bufferpop=new QPopupMenu;
  id_buffer=menubar->insertItem("&Buffer", bufferpop,-2);
  connect(bufferpop,SIGNAL(activated(int)),this,SLOT(buffersel(int)));

  tasks = new QPopupMenu;
  id_plic=tasks->insertItem("&Pascalli->Pascallino",this,SLOT(pascalli2pascallino()));
  id_ploc=tasks->insertItem("Pascallino->&Ram",this,SLOT(pascallino2ram()));
  id_ramsim=tasks->insertItem("Ram&sim",this,SLOT(ramsim()));
  id_tasks=menubar->insertItem("&Tasks",tasks,-2);

  menubar->insertSeparator();

  QPopupMenu *help = new QPopupMenu;
  help->insertItem( "&About", this, SLOT(about()) );
  help->insertItem( "About &Qt", this, SLOT(aboutQt()) );
  menubar->insertItem("&Help", help );
  v->setMenuBar( menubar );

  editstack=new QWidgetStack(this);
  bgwdg *b = new bgwdg(editstack);

  editstack->setMinimumSize(b->size());
  // stupid QT. this one is needed on qt 2.0.2 but not qt1.4 and not qt2.1

  editstack->addWidget(b,-10);
  v->addWidget(editstack,1);
  buffersel(-10);

  v->activate();
};

void mainw::buffersel (int id)
{
  menubar->setItemEnabled(id_buffer, bufferpop->count());

  if (id==-10) {
    // bgwidget
    menubar->setItemEnabled(id_tasks,FALSE);
    filepop->setItemEnabled(id_file_save,FALSE);
    filepop->setItemEnabled(id_file_close,FALSE);
    tasks->setItemEnabled(id_plic,FALSE);
    tasks->setItemEnabled(id_ploc,FALSE);
    tasks->setItemEnabled(id_ramsim,FALSE);
  }else{
    // a real buffer
    menubar->setItemEnabled(id_tasks,TRUE);
    filepop->setItemEnabled(id_file_save,TRUE);
    filepop->setItemEnabled(id_file_close,TRUE);
    buffer *b=static_cast<buffer*>(editstack->widget(id));
    tasks->setItemEnabled(id_plic,b->filetype==PASCALLI);
    tasks->setItemEnabled(id_ploc,b->filetype==PASCALLINO);
    tasks->setItemEnabled(id_ramsim,b->filetype==RAM);
  };
  editstack->raiseWidget(id);
};

void mainw::fileopen()
{
  // FIXME this does display Save if you want to open.
  QFileDialog *fd = new QFileDialog(this,0,TRUE);
  QString filename;
  fd->setMode(QFileDialog::AnyFile);
  if (fd->exec()) filename=fd->selectedFile();

  if (!filename.isNull()) {
    if (findbuffer(filename)) {
      buffersel(editstack->id(findbuffer(filename)));
    }else{
      realopen(filename);
    };
  };
};

void mainw::filesave()
{
  buffer *visible=(buffer*)editstack->visibleWidget();
  QFile f( visible->filename );
  if ( !f.open( IO_WriteOnly ) ) fatal("cannot open file for writing");
  
  QTextStream t(&f);
  for (int i=0;i<visible->editor->numLines();i++) {
    t<<visible->editor->textLine(i);
    t<<'\n';
  };
  f.close();

  visible->changed=false;
  visible->updatestatus();
};

void mainw::fileclose()
{
  buffer *visible=(buffer*) editstack->visibleWidget();
  int id=editstack->id(visible);

  if (visible->changed) {
    switch( QMessageBox::information( this, "theosim",
                                      "The document contains unsaved work\n"
                                      "Do you want to save it before closing?",
                                      "&Save", "&Don't Save", "&Cancel",
                                      0,      // Enter == button 0
                                      2 ) ) { // Escape == button 2
    case 0: // Save clicked, Alt-S or Enter pressed.
      filesave();
      break;
    case 1: // Don't Save clicked or Alt-D pressed
      break;
    case 2: // Cancel clicked, Alt-C or Escape pressed
      return;
      break;
    }
  };

  buffersel(-10);
  bufferpop->removeItem(id);
  editstack->removeWidget(visible);
  delete visible;
  buffersel(-10);
  if (bufferpop->count()) buffersel(bufferpop->idAt(0));
};

void mainw::filequit()
{
  while (bufferpop->count()) {
    buffersel(bufferpop->idAt(0));

    buffer *visible=(buffer*) editstack->visibleWidget();
    if (visible->changed) {
      switch( QMessageBox::information( this, "theosim",
					"The document contains unsaved work\n"
					"Do you want to save it before closing?",
					"&Save", "&Don't Save", "&Cancel",
					0,      // Enter == button 0
					2 ) ) { // Escape == button 2
      case 0: // Save clicked, Alt-S or Enter pressed.
	filesave();
	break;
      case 1: // Don't Save clicked or Alt-D pressed
	visible->changed=false;
	break;
      case 2: // Cancel clicked, Alt-C or Escape pressed
	return;
	break;
      };
    };
    fileclose();
  };
  
  qApp->quit();
};

void mainw::pascalli2pascallino()
{
  filesave();
  buffer *visible=(buffer*)editstack->visibleWidget();

  QString base(visible->filename);
  base.truncate(QRegExp(".pli$").match(base));

  QString source=base+".pli";
  QString destin=base+".plo";

  buffer *b=findbuffer(destin);
  if (b) {
    buffersel(editstack->id(b));
    fileclose();
    if (findbuffer(destin)) return;
  };
  buffersel(editstack->id(findbuffer(source)));

  char *src=(char*)malloc(source.length()+1);
  memcpy(src,static_cast<const char *const>(source), source.length());
  src[source.length()]=0;

  char *dst=(char*)malloc(destin.length()+1);
  memcpy(dst,static_cast<const char *const>(destin), destin.length());
  dst[destin.length()]=0;

  char *const args[] = { "plic", src, dst, 0 };
  external *e = new external("plic", args[0], args, this);
  e->exec();

  free(dst);
  free(src);

  if (e->result==external::success) realopen(destin);
};

void mainw::pascallino2ram()
{
  filesave();
  buffer *visible=(buffer*)editstack->visibleWidget();

  QString base(visible->filename);
  base.truncate(QRegExp(".plo$").match(base));

  QString source=base+".plo";
  QString destin=base+".ram";

  buffer *b=findbuffer(destin);
  if (b) {
    buffersel(editstack->id(b));
    fileclose();
    if (findbuffer(destin)) return;
  };
  buffersel(editstack->id(findbuffer(source)));

  char *src=(char*)malloc(source.length()+1);
  memcpy(src,static_cast<const char *const>(source), source.length());
  src[source.length()]=0;

  char *dst=(char*)malloc(destin.length()+1);
  memcpy(dst,static_cast<const char *const>(destin), destin.length());
  dst[destin.length()]=0;

  char *const args[] = { "ploc", src, dst, 0 };
  external *e = new external("ploc", args[0], args, this);
  e->exec();

  free(dst);
  free(src);

  if (e->result==external::success) realopen(destin);
};

void mainw::ramsim()
{
  filesave();
  buffer *visible=(buffer*)editstack->visibleWidget();
  char *fn=(char*)malloc(visible->filename.length()+1);
  memcpy(fn,static_cast<const char *const>(visible->filename),visible->filename.length());
  fn[visible->filename.length()]=0;
  char *const args[] = { "ramsim", fn, 0 };
  external *e = new external("ramsim", args[0], args, this);
  e->exec();
  free(fn);
};

void mainw::about()
{
  QString msg;
  msg+="theosim version "; msg+=VERSION; msg+=", Copyright (C) 1999  Erik Thiele\n";
  msg+="theosim comes with ABSOLUTELY NO WARRANTY; for details read file LICENSE\n"
    "This is free software, and you are welcome to redistribute it\n"
    "under certain conditions; for details read file LICENSE\n"
    "\n"
    "theosim is a GUI to integrate the\n"
    "Pascalli -> Pascallino compiler (plic), the\n"
    "Pascallino -> RAM compiler (ploc), and the\n"
    "RAM simulator (ramsim)\n"
    "in one single program";
  QMessageBox::about( this, "theosim", msg );
};

void mainw::aboutQt()
{
  QMessageBox::aboutQt( this, "theosim" );
};

void mainw::closeEvent( QCloseEvent *e )
{
  filequit();
  e->ignore();
};

buffer *mainw::findbuffer (const QString filename)
{
  for (unsigned i=0;i<bufferpop->count();i++) {
    int id=bufferpop->idAt(i);
    buffer *b=static_cast<buffer*>(editstack->widget(id));
    if (b->filename==filename) return b;
  };
  return 0;
};

void mainw::realopen(QString filename)
{
  if (!filename.isNull()) {
    buffer *neu = new buffer(editstack);

    if (QRegExp(".pli$").match(filename)!=-1) neu->filetype=PASCALLI;
    else if (QRegExp(".plo$").match(filename)!=-1) neu->filetype=PASCALLINO;
    else if (QRegExp(".ram$").match(filename)!=-1) neu->filetype=RAM;
    else {
      delete neu;
      QMessageBox::critical(this,"Error","Filename must end with\n.pli .plo or .ram !");
      return;
    };

    neu->editor->setAutoUpdate( FALSE );
    neu->filename=filename;
    editstack->addWidget(neu,bufferpop->insertItem(neu->filename,-2));
    buffersel(editstack->id(neu));
    neu->editor->setFocus();
    
    QFile f( neu->filename );
    if ( f.open( IO_ReadOnly ) ) {
      QTextStream t(&f);
      while ( !t.eof() ) {
	QString s = t.readLine();
	neu->editor->append( s );
      }
      f.close();
      neu->changed=FALSE;
    }else{
      neu->changed=TRUE;
    };

    neu->updatestatus();
    neu->editor->setAutoUpdate( TRUE );
    neu->editor->repaint();
  };
};

QPixmap *pixmap_theosim;
